###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", fWITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################
import base64
import binascii
import inspect
import math
import os
import random
import re
import socket
import struct
import subprocess
import sys
import time
from array import array
from collections import OrderedDict
from datetime import datetime, timedelta
from pprint import pformat

import txaio

try:
    _TLS = True
    from OpenSSL import SSL
except ImportError:
    _TLS = False


__all__ = (
    "EqualityMixin",
    "IdGenerator",
    "ObservableMixin",
    "Stopwatch",
    "Tracker",
    "encode_truncate",
    "generate_activation_code",
    "generate_serial_number",
    "generate_token",
    "generate_user_password",
    "hl",
    "hlcontract",
    "hlid",
    "hltype",
    "hluserid",
    "hlval",
    "id",
    "machine_id",
    "newid",
    "parse_keyfile",
    "public",
    "rid",
    "rtime",
    "utcnow",
    "utcstr",
    "with_0x",
    "without_0x",
    "write_keyfile",
    "xor",
)


def public(obj):
    """
    The public user API of Autobahn is marked using this decorator.
    Everything that is not decorated @public is library internal, can
    change at any time and should not be used in user program code.
    """
    try:
        obj._is_public = True
    except AttributeError:
        # FIXME: exceptions.AttributeError: 'staticmethod' object has no attribute '_is_public'
        pass
    return obj


@public
def encode_truncate(text, limit, encoding="utf8", return_encoded=True):
    """
    Given a string, return a truncated version of the string such that
    the UTF8 encoding of the string is smaller than the given limit.

    This function correctly truncates even in the presence of Unicode code
    points that encode to multi-byte encodings which must not be truncated
    in the middle.

    :param text: The (Unicode) string to truncate.
    :type text: str
    :param limit: The number of bytes to limit the UTF8 encoding to.
    :type limit: int
    :param encoding: Truncate the string in this encoding (default is ``utf-8``).
    :type encoding: str
    :param return_encoded: If ``True``, return the string encoded into bytes
        according to the specified encoding, else return the string as a string.
    :type return_encoded: bool

    :returns: The truncated string.
    :rtype: str or bytes
    """
    assert text is None or type(text) == str
    assert type(limit) == int
    assert limit >= 0

    if text is None:
        return

    # encode the given string in the specified encoding
    s = text.encode(encoding)

    # when the resulting byte string is longer than the given limit ..
    if len(s) > limit:
        # .. truncate, and
        s = s[:limit]

        # decode back, ignoring errors that result from truncation
        # in the middle of multi-byte encodings
        text = s.decode(encoding, "ignore")

        if return_encoded:
            s = text.encode(encoding)

    if return_encoded:
        return s
    else:
        return text


@public
def xor(d1: bytes, d2: bytes) -> bytes:
    """
    XOR two binary strings of arbitrary (equal) length.

    :param d1: The first binary string.
    :param d2: The second binary string.

    :returns: XOR of the binary strings (``XOR(d1, d2)``)
    """
    if type(d1) != bytes:
        raise Exception(f"invalid type {type(d1)} for d1 - must be binary")
    if type(d2) != bytes:
        raise Exception(f"invalid type {type(d2)} for d2 - must be binary")
    if len(d1) != len(d2):
        raise Exception(
            f"cannot XOR binary string of differing length ({len(d1)} != {len(d2)})"
        )

    d1 = array("B", d1)
    d2 = array("B", d2)

    for i in range(len(d1)):
        d1[i] ^= d2[i]

    return d1.tobytes()


@public
def utcstr(ts=None):
    """
    Format UTC timestamp in ISO 8601 format.

    Note: to parse an ISO 8601 formatted string, use the **iso8601**
    module instead (e.g. ``iso8601.parse_date("2014-05-23T13:03:44.123Z")``).

    >>> txaio.time_ns()
    1641121311914026419
    >>> int(iso8601.parse_date(utcnow()).timestamp() * 1000000000.)
    1641121313209000192

    :param ts: The timestamp to format.
    :type ts: instance of :py:class:`datetime.datetime` or ``None``

    :returns: Timestamp formatted in ISO 8601 format.
    :rtype: str
    """
    assert ts is None or isinstance(ts, datetime)
    if ts is None:
        ts = datetime.utcnow()
    return "{}Z".format(ts.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3])


@public
def utcnow():
    """
    Get current time in UTC as ISO 8601 string.

    :returns: Current time as string in ISO 8601 format.
    :rtype: str
    """
    return utcstr()


class IdGenerator:
    """
    ID generator for WAMP request IDs.

    WAMP request IDs are sequential per WAMP session, starting at 1 and
    wrapping around at 2**53 (both value are inclusive [1, 2**53]).

    The upper bound **2**53** is chosen since it is the maximum integer that can be
    represented as a IEEE double such that all smaller integers are representable as well.

    Hence, IDs can be safely used with languages that use IEEE double as their
    main (or only) number type (JavaScript, Lua, etc).

    See https://github.com/wamp-proto/wamp-proto/blob/master/spec/basic.md#ids
    """

    def __init__(self):
        self._next = 0  # starts at 1; next() pre-increments

    def next(self):
        """
        Returns next ID.

        :returns: The next ID.
        :rtype: int
        """
        self._next += 1
        if self._next > 9007199254740992:
            self._next = 1
        return self._next

    # generator protocol
    def __next__(self):
        return self.next()


#
# Performance comparison of IdGenerator.next(), id() and rid().
#
# All tests were performed on:
#
#   - Ubuntu 14.04 LTS x86-64
#   - Intel Core i7 920 @ 3.3GHz
#
# The tests generated 100 mio. IDs and run-time was measured
# as wallclock from Unix "time" command. In each run, a single CPU
# core was essentially at 100% load all the time (though the sys/usr
# ratio was different).
#
# PyPy 2.6.1:
#
#   IdGenerator.next()    0.5s
#   id()                 29.4s
#   rid()               106.1s
#
# CPython 2.7.10:
#
#   IdGenerator.next()   49.0s
#   id()                370.5s
#   rid()               196.4s
#

#
# Note on the ID range [0, 2**53]. We once reduced the range to [0, 2**31].
# This lead to extremely hard to track down issues due to ID collisions!
# Here: https://github.com/crossbario/autobahn-python/issues/419#issue-90483337
#


# 8 byte mask with 53 LSBs set (WAMP requires IDs from [1, 2**53]
_WAMP_ID_MASK = struct.unpack(">Q", b"\x00\x1f\xff\xff\xff\xff\xff\xff")[0]


def rid():
    """
    Generate a new random integer ID from range **[1, 2**53]**.

    The generated ID is uniformly distributed over the whole range, doesn't have
    a period (no pseudo-random generator is used) and cryptographically strong.

    The upper bound **2**53** is chosen since it is the maximum integer that can be
    represented as a IEEE double such that all smaller integers are representable as well.

    Hence, IDs can be safely used with languages that use IEEE double as their
    main (or only) number type (JavaScript, Lua, etc).

    :returns: A random integer ID.
    :rtype: int
    """
    return struct.unpack(">Q", os.urandom(8))[0] & _WAMP_ID_MASK or 2**53


# noinspection PyShadowingBuiltins
def id():
    """
    Generate a new random integer ID from range **[1, 2**53]**.

    The generated ID is based on a pseudo-random number generator (Mersenne Twister,
    which has a period of 2**19937-1). It is NOT cryptographically strong, and
    hence NOT suitable to generate e.g. secret keys or access tokens.

    The upper bound **2**53** is chosen since it is the maximum integer that can be
    represented as a IEEE double such that all smaller integers are representable as well.

    Hence, IDs can be safely used with languages that use IEEE double as their
    main (or only) number type (JavaScript, Lua, etc).

    :returns: A random integer ID.
    :rtype: int
    """
    return random.randint(1, 9007199254740992)


def newid(length=16):
    """
    Generate a new random string ID.

    The generated ID is uniformly distributed and cryptographically strong. It is
    hence usable for things like secret keys and access tokens.

    :param length: The length (in chars) of the ID to generate.
    :type length: int

    :returns: A random string ID.
    :rtype: str
    """
    l = int(math.ceil(float(length) * 6.0 / 8.0))
    return base64.b64encode(os.urandom(l))[:length].decode("ascii")


# a standard base36 character set
# DEFAULT_TOKEN_CHARS = string.digits + string.ascii_uppercase

# we take out the following 9 chars (leaving 27), because there
# is visual ambiguity: 0/O/D, 1/I, 8/B, 2/Z
DEFAULT_TOKEN_CHARS = "345679ACEFGHJKLMNPQRSTUVWXY"
"""
Default set of characters to create rtokens from.
"""

DEFAULT_ZBASE32_CHARS = "13456789abcdefghijkmnopqrstuwxyz"
"""
Our choice of confusing characters to eliminate is: `0', `l', `v', and `2'.  Our
reasoning is that `0' is potentially mistaken for `o', that `l' is potentially
mistaken for `1' or `i', that `v' is potentially mistaken for `' or `r'
(especially in handwriting) and that `2' is potentially mistaken for `z'
(especially in handwriting).

Note that we choose to focus on typed and written transcription more than on
vocal, since humans already have a well-established system of disambiguating
spoken alphanumerics, such as the United States military's "Alpha Bravo Charlie
Delta" and telephone operators' "Is that 'd' as in 'dog'?".

* http://philzimmermann.com/docs/human-oriented-base-32-encoding.txt
"""


@public
def generate_token(
    char_groups: int,
    chars_per_group: int,
    chars: str | None = None,
    sep: str | None = None,
    lower_case: bool | None = False,
) -> str:
    """
    Generate cryptographically strong tokens, which are strings like `M6X5-YO5W-T5IK`.
    These can be used e.g. for used-only-once activation tokens or the like.

    The returned token has an entropy of
    ``math.log(len(chars), 2.) * chars_per_group * char_groups``
    bits.

    With the default charset and 4 characters per group, ``generate_token()`` produces
    strings with the following entropy:

    ================   ===================  ========================================
    character groups    entropy (at least)  recommended use
    ================   ===================  ========================================
    2                    38 bits
    3                    57 bits            one-time activation or pairing code
    4                    76 bits            secure user password
    5                    95 bits
    6                   114 bits            globally unique serial / product code
    7                   133 bits
    ================   ===================  ========================================

    Here are some examples:

    * token(3): ``9QXT-UXJW-7R4H``
    * token(4): ``LPNN-JMET-KWEP-YK45``
    * token(6): ``NXW9-74LU-6NUH-VLPV-X6AG-QUE3``

    :param char_groups: Number of character groups (or characters if chars_per_group == 1).
    :param chars_per_group: Number of characters per character group (or 1 to return a token with no grouping).
    :param chars: Characters to choose from. Default is 27 character subset
        of the ISO basic Latin alphabet (see: ``DEFAULT_TOKEN_CHARS``).
    :param sep: When separating groups in the token, the separater string.
    :param lower_case: If ``True``, generate token in lower-case.

    :returns: The generated token.
    """
    assert type(char_groups) == int
    assert type(chars_per_group) == int
    assert chars is None or type(chars) == str, f"chars must be str, was {type(chars)}"
    chars = chars or DEFAULT_TOKEN_CHARS
    if lower_case:
        chars = chars.lower()
    sep = sep or "-"
    rng = random.SystemRandom()
    token_value = "".join(
        rng.choice(chars) for _ in range(char_groups * chars_per_group)
    )
    if chars_per_group > 1:
        return sep.join(map("".join, zip(*[iter(token_value)] * chars_per_group)))
    else:
        return token_value


@public
def generate_activation_code():
    """
    Generate a one-time activation code or token of the form ``'W97F-96MJ-YGJL'``.
    The generated value is cryptographically strong and has (at least) 57 bits of entropy.

    :returns: The generated activation code.
    :rtype: str
    """
    return generate_token(
        char_groups=3,
        chars_per_group=4,
        chars=DEFAULT_TOKEN_CHARS,
        sep="-",
        lower_case=False,
    )


_PAT_ACTIVATION_CODE = re.compile(
    "^(["
    + DEFAULT_TOKEN_CHARS
    + "]{4,4})-(["
    + DEFAULT_TOKEN_CHARS
    + "]{4,4})-(["
    + DEFAULT_TOKEN_CHARS
    + "]{4,4})$"
)


@public
def parse_activation_code(code: str):
    """
    Parse an activation code generated by :func:<autobahn.util.generate_activation_code>:

    .. code:: console

        "RWCN-94NV-CEHR" -> ("RWCN", "94NV", "CEHR") | None

    :param code: The code to parse, e.g. ``'W97F-96MJ-YGJL'``.
    :return: If the string is a properly conforming activation code, return
        the matched pattern, otherwise return ``None``.
    """
    return _PAT_ACTIVATION_CODE.match(code)


@public
def generate_user_password():
    """
    Generate a secure, random user password of the form ``'kgojzi61dn5dtb6d'``.
    The generated value is cryptographically strong and has (at least) 76 bits of entropy.

    :returns: The generated password.
    :rtype: str
    """
    return generate_token(
        char_groups=16,
        chars_per_group=1,
        chars=DEFAULT_ZBASE32_CHARS,
        sep="-",
        lower_case=True,
    )


@public
def generate_serial_number():
    """
    Generate a globally unique serial / product code of the form ``'YRAC-EL4X-FQQE-AW4T-WNUV-VN6T'``.
    The generated value is cryptographically strong and has (at least) 114 bits of entropy.

    :returns: The generated serial number / product code.
    :rtype: str
    """
    return generate_token(
        char_groups=6,
        chars_per_group=4,
        chars=DEFAULT_TOKEN_CHARS,
        sep="-",
        lower_case=False,
    )


# Select the most precise walltime measurement function available
# on the platform
#
if sys.platform.startswith("win"):
    # On Windows, this function returns wall-clock seconds elapsed since the
    # first call to this function, as a floating point number, based on the
    # Win32 function QueryPerformanceCounter(). The resolution is typically
    # better than one microsecond
    _rtime = time.perf_counter
    _ = _rtime()  # this starts wallclock
else:
    # On Unix-like platforms, this used the first available from this list:
    # (1) gettimeofday() -- resolution in microseconds
    # (2) ftime() -- resolution in milliseconds
    # (3) time() -- resolution in seconds
    _rtime = time.time


@public
def rtime():
    """
    Precise, fast wallclock time.

    :returns: The current wallclock in seconds. Returned values are only guaranteed
       to be meaningful relative to each other.
    :rtype: float
    """
    return _rtime()


class Stopwatch:
    """
    Stopwatch based on walltime.

    This can be used to do code timing and uses the most precise walltime measurement
    available on the platform. This is a very light-weight object,
    so create/dispose is very cheap.
    """

    def __init__(self, start=True):
        """

        :param start: If ``True``, immediately start the stopwatch.
        :type start: bool
        """
        self._elapsed = 0
        if start:
            self._started = rtime()
            self._running = True
        else:
            self._started = None
            self._running = False

    def elapsed(self):
        """
        Return total time elapsed in seconds during which the stopwatch was running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        if self._running:
            now = rtime()
            return self._elapsed + (now - self._started)
        else:
            return self._elapsed

    def pause(self):
        """
        Pauses the stopwatch and returns total time elapsed in seconds during which
        the stopwatch was running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        if self._running:
            now = rtime()
            self._elapsed += now - self._started
            self._running = False
            return self._elapsed
        else:
            return self._elapsed

    def resume(self):
        """
        Resumes a paused stopwatch and returns total elapsed time in seconds
        during which the stopwatch was running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        if not self._running:
            self._started = rtime()
            self._running = True
            return self._elapsed
        else:
            now = rtime()
            return self._elapsed + (now - self._started)

    def stop(self):
        """
        Stops the stopwatch and returns total time elapsed in seconds during which
        the stopwatch was (previously) running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        elapsed = self.pause()
        self._elapsed = 0
        self._started = None
        self._running = False
        return elapsed


class Tracker:
    """
    A key-based statistics tracker.
    """

    def __init__(self, tracker, tracked):
        """ """
        self.tracker = tracker
        self.tracked = tracked
        self._timings = {}
        self._offset = rtime()
        self._dt_offset = datetime.utcnow()

    def track(self, key):
        """
        Track elapsed for key.

        :param key: Key under which to track the timing.
        :type key: str
        """
        self._timings[key] = rtime()

    def diff(self, start_key, end_key, formatted=True):
        """
        Get elapsed difference between two previously tracked keys.

        :param start_key: First key for interval (older timestamp).
        :type start_key: str
        :param end_key: Second key for interval (younger timestamp).
        :type end_key: str
        :param formatted: If ``True``, format computed time period and return string.
        :type formatted: bool

        :returns: Computed time period in seconds (or formatted string).
        :rtype: float or str
        """
        if end_key in self._timings and start_key in self._timings:
            d = self._timings[end_key] - self._timings[start_key]
            if formatted:
                if d < 0.00001:  # 10us
                    s = f"{round(d * 1000000000.0)} ns"
                elif d < 0.01:  # 10ms
                    s = f"{round(d * 1000000.0)} us"
                elif d < 10:  # 10s
                    s = f"{round(d * 1000.0)} ms"
                else:
                    s = f"{round(d)} s"
                return s.rjust(8)
            else:
                return d
        else:
            if formatted:
                return "n.a.".rjust(8)
            else:
                return None

    def absolute(self, key):
        """
        Return the UTC wall-clock time at which a tracked event occurred.

        :param key: The key
        :type key: str

        :returns: Timezone-naive datetime.
        :rtype: instance of :py:class:`datetime.datetime`
        """
        elapsed = self[key]
        if elapsed is None:
            raise KeyError(f'No such key "{elapsed}".')
        return self._dt_offset + timedelta(seconds=elapsed)

    def __getitem__(self, key):
        if key in self._timings:
            return self._timings[key] - self._offset
        else:
            return None

    def __iter__(self):
        return self._timings.__iter__()

    def __str__(self):
        return pformat(self._timings)


class EqualityMixin:
    """
    Mixing to add equality comparison operators to a class.

    Two objects are identical under this mixin, if and only if:

    1. both object have the same class
    2. all non-private object attributes are equal
    """

    def __eq__(self, other):
        """
        Compare this object to another object for equality.

        :param other: The other object to compare with.
        :type other: obj

        :returns: ``True`` iff the objects are equal.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        # we only want the actual message data attributes (not eg _serialize)
        for k in self.__dict__:
            if not k.startswith("_"):
                if not self.__dict__[k] == other.__dict__[k]:
                    return False
        return True
        # return (isinstance(other, self.__class__) and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        """
        Compare this object to another object for inequality.

        :param other: The other object to compare with.
        :type other: obj

        :returns: ``True`` iff the objects are not equal.
        :rtype: bool
        """
        return not self.__eq__(other)


def wildcards2patterns(wildcards):
    """
    Compute a list of regular expression patterns from a list of
    wildcard strings. A wildcard string uses '*' as a wildcard character
    matching anything.

    :param wildcards: List of wildcard strings to compute regular expression patterns for.
    :type wildcards: list of str

    :returns: Computed regular expressions.
    :rtype: list of obj
    """
    # note that we add the ^ and $ so that the *entire* string must
    # match. Without this, e.g. a prefix will match:
    # re.match('.*good\\.com', 'good.com.evil.com')  # match!
    # re.match('.*good\\.com$', 'good.com.evil.com') # no match!
    return [
        re.compile("^" + wc.replace(".", r"\.").replace("*", ".*") + "$")
        for wc in wildcards
    ]


class ObservableMixin:
    """
    Internal utility for enabling event-listeners on particular objects
    """

    # A "helper" style composable class (as opposed to a mix-in) might
    # be a lot easier to deal with here.  Having an __init__ method
    # with a "mix in" style class can be fragile and error-prone,
    # especially if it takes arguments. Since we don't use the
    # "parent" beavior anywhere, I didn't add a .set_parent() (yet?)

    # these are class-level globals; individual instances are
    # initialized as-needed (e.g. the first .on() call adds a
    # _listeners dict). Thus, subclasses don't have to call super()
    # properly etc.
    _parent = None
    _valid_events = None
    _listeners = None
    _results = None

    def set_valid_events(self, valid_events=None):
        """
        :param valid_events: if non-None, .on() or .fire() with an event
            not listed in valid_events raises an exception.
        """
        self._valid_events = list(valid_events)
        self._results = {k: None for k in self._valid_events}

    def _check_event(self, event):
        """
        Internal helper. Throws RuntimeError if we have a valid_events
        list, and the given event isnt' in it. Does nothing otherwise.
        """
        if self._valid_events and event not in self._valid_events:
            raise RuntimeError(
                "Invalid event '{event}'. Expected one of: {events}".format(
                    event=event,
                    events=", ".join(self._valid_events),
                )
            )

    def on(self, event, handler):
        """
        Add a handler for an event.

        :param event: the name of the event

        :param handler: a callable thats invoked when .fire() is
            called for this events. Arguments will be whatever are given
            to .fire()
        """
        # print("adding '{}' to '{}': {}".format(event, hash(self), handler))
        self._check_event(event)
        if self._listeners is None:
            self._listeners = dict()
        if event not in self._listeners:
            self._listeners[event] = []
        self._listeners[event].append(handler)

    def off(self, event=None, handler=None):
        """
        Stop listening for a single event, or all events.

        :param event: if None, remove all listeners. Otherwise, remove
            listeners for the single named event.

        :param handler: if None, remove all handlers for the named
            event; otherwise remove just the given handler.
        """
        if event is None:
            if handler is not None:
                # maybe this should mean "remove the given handler
                # from any event at all that contains it"...?
                raise RuntimeError(
                    "Can't specificy a specific handler without an event"
                )
            self._listeners = dict()
        else:
            if self._listeners is None:
                return
            self._check_event(event)
            if event in self._listeners:
                if handler is None:
                    del self._listeners[event]
                else:
                    try:
                        self._listeners[event].remove(handler)
                    except ValueError:
                        pass

    def fire(self, event, *args, **kwargs):
        """
        Fire a particular event.

        :param event: the event to fire. All other args and kwargs are
            passed on to the handler(s) for the event.

        :return: a Deferred/Future gathering all async results from
            all handlers and/or parent handlers.
        """
        # print("firing '{}' from '{}'".format(event, hash(self)))
        if self._listeners is None:
            return txaio.create_future(result=[])

        self._check_event(event)
        res = []
        for handler in self._listeners.get(event, []):
            future = txaio.as_future(handler, *args, **kwargs)
            res.append(future)
        if self._parent is not None:
            res.append(self._parent.fire(event, *args, **kwargs))
        d_res = txaio.gather(res, consume_exceptions=False)
        self._results[event] = d_res
        return d_res


class _LazyHexFormatter:
    """
    This is used to avoid calling binascii.hexlify() on data given to
    log.debug() calls unless debug is active (for example). Like::

        self.log.debug(
            "Some data: {octets}",
            octets=_LazyHexFormatter(os.urandom(32)),
        )
    """

    __slots__ = ("obj",)

    def __init__(self, obj):
        self.obj = obj

    def __str__(self):
        return binascii.hexlify(self.obj).decode("ascii")


def _is_tls_error(instance):
    """
    :returns: True if we have TLS support and 'instance' is an
        instance of :class:`OpenSSL.SSL.Error` otherwise False
    """
    if _TLS:
        return isinstance(instance, SSL.Error)
    return False


def _maybe_tls_reason(instance):
    """
    :returns: a TLS error-message, or empty-string if 'instance' is
        not a TLS error.
    """
    if _is_tls_error(instance):
        ssl_error = instance.args[0][0]
        return f"SSL error: {ssl_error[2]} (in {ssl_error[1]})"
    return ""


def machine_id() -> str:
    """
    For informational purposes, get a unique ID or serial for this machine (device).

    :returns: Unique machine (device) ID (serial), e.g. ``81655b901e334fc1ad59cbf2719806b7``.
    """
    from twisted.python.runtime import platform

    if platform.isLinux():
        try:
            # why this? see: http://0pointer.de/blog/projects/ids.html
            with open("/var/lib/dbus/machine-id") as f:
                return f.read().strip()
        except:
            # Non-dbus using Linux, get a hostname
            return socket.gethostname()
    elif platform.isMacOSX():
        import plistlib

        plist_data = subprocess.check_output(
            ["ioreg", "-rd1", "-c", "IOPlatformExpertDevice", "-a"]
        )
        return plistlib.loads(plist_data)[0]["IOPlatformSerialNumber"]
    else:
        return socket.gethostname()


try:
    import click

    _HAS_CLICK = True
except ImportError:
    _HAS_CLICK = False


def hl(text, bold=False, color="yellow"):
    if not isinstance(text, str):
        text = f"{text}"
    if _HAS_CLICK:
        return click.style(text, fg=color, bold=bold)
    else:
        return text


def _qn(obj):
    if inspect.isclass(obj) or inspect.isfunction(obj) or inspect.ismethod(obj):
        qn = f"{obj.__module__}.{obj.__qualname__}"
    else:
        qn = "unknown"
    return qn


def hltype(obj):
    qn = _qn(obj).split(".")
    text = hl(qn[0], color="yellow", bold=True) + hl(
        "." + ".".join(qn[1:]), color="yellow", bold=False
    )
    return "<" + text + ">"


def hlid(oid):
    return hl(f"{oid}", color="blue", bold=True)


def hluserid(oid):
    if not isinstance(oid, str):
        oid = f"{oid}"
    return hl(f'"{oid}"', color="yellow", bold=True)


def hlval(val, color="white", bold=True):
    return hl(f"{val}", color=color, bold=bold)


def hlcontract(oid):
    if not isinstance(oid, str):
        oid = f"{oid}"
    return hl(f"<{oid}>", color="magenta", bold=True)


def with_0x(address):
    if address and not address.startswith("0x"):
        return f"0x{address}"
    return address


def without_0x(address):
    if address and address.startswith("0x"):
        return address[2:]
    return address


def write_keyfile(filepath, tags, msg):
    """
    Internal helper, write the given tags to the given file-
    """
    with open(filepath, "w") as f:
        f.write(msg)
        for tag, value in tags.items():
            if value:
                f.write(f"{tag}: {value}\n")


def parse_keyfile(key_path: str, private: bool = True) -> OrderedDict:
    """
    Internal helper. This parses a node.pub or node.priv file and
    returns a dict mapping tags -> values.
    """
    if os.path.exists(key_path) and not os.path.isfile(key_path):
        raise Exception(f"Key file '{key_path}' exists, but isn't a file")

    allowed_tags = [
        # common tags
        "public-key-ed25519",
        "public-adr-eth",
        "created-at",
        "creator",
        # user profile
        "user-id",
        # node profile
        "machine-id",
        "node-authid",
        "node-cluster-ip",
    ]

    if private:
        # private key file tags
        allowed_tags.extend(["private-key-ed25519", "private-key-eth"])

    tags = OrderedDict()  # type: ignore
    with open(key_path) as key_file:
        got_blankline = False
        for line in key_file.readlines():
            if line.strip() == "":
                got_blankline = True
            elif got_blankline:
                tag, value = line.split(":", 1)
                tag = tag.strip().lower()
                value = value.strip()
                if tag not in allowed_tags:
                    raise Exception(
                        f"Invalid tag '{tag}' in key file {key_path}"
                    )
                if tag in tags:
                    raise Exception(
                        f"Duplicate tag '{tag}' in key file {key_path}"
                    )
                tags[tag] = value
    return tags



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP Payload Passthru (PPT) application payload cipher (NEW!).
class PPTCipher(object):
    # No valid cipher (unfilled)
    NULL = 0
    # Particular combination of Salsa20 and Poly1305 specified in Daniel J. Bernstein, "Cryptography in NaCl" (https://cr.yp.to/highspeed/naclcrypto-20090310.pdf), see NaCl (https://nacl.cr.yp.to/) "crypto_secretbox_xsalsa20poly1305" (aka WAMP "cryptobox").
    XSALSA20POLY1305 = 1
    # Galois/Counter Mode (GCM), see https://en.wikipedia.org/wiki/Galois/Counter_Mode.
    AES256GCM = 2



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import annotations

from base64 import b64decode, b64encode
from typing import Any

import txaio
from zope.interface import implementer

txaio.use_twisted()

import twisted.internet.protocol
from autobahn.twisted.util import create_transport_details, transport_channel_id
from autobahn.util import _is_tls_error, _maybe_tls_reason, hltype, hlval, public
from autobahn.wamp import websocket
from autobahn.wamp.types import TransportDetails
from autobahn.websocket import protocol
from autobahn.websocket.compress import (
    PerMessageDeflateOffer,
    PerMessageDeflateOfferAccept,
    PerMessageDeflateResponse,
    PerMessageDeflateResponseAccept,
)
from autobahn.websocket.interfaces import IWebSocketClientAgent
from autobahn.websocket.types import (
    ConnectionDeny,
    ConnectionRequest,
    ConnectionResponse,
)
from twisted.internet import endpoints
from twisted.internet.defer import Deferred
from twisted.internet.error import ConnectionAborted, ConnectionDone, ConnectionLost
from twisted.internet.interfaces import ITransport
from twisted.internet.protocol import connectionDone
from twisted.python.failure import Failure

__all__ = (
    "WampWebSocketClientFactory",
    "WampWebSocketClientProtocol",
    "WampWebSocketServerFactory",
    "WampWebSocketServerProtocol",
    "WebSocketAdapterFactory",
    "WebSocketAdapterProtocol",
    "WebSocketClientFactory",
    "WebSocketClientProtocol",
    "WebSocketServerFactory",
    "WebSocketServerProtocol",
    "WrappingWebSocketAdapter",
    "WrappingWebSocketClientFactory",
    "WrappingWebSocketClientProtocol",
    "WrappingWebSocketServerFactory",
    "WrappingWebSocketServerProtocol",
    "connectWS",
    "create_client_agent",
    "listenWS",
)


def create_client_agent(reactor) -> _TwistedWebSocketClientAgent:
    """
    :returns: an instance implementing IWebSocketClientAgent
    """
    return _TwistedWebSocketClientAgent(reactor)


def check_transport_config(transport_config: str) -> None:
    """
    raises a ValueError if `transport_config` is invalid
    """
    # XXX move me to "autobahn.websocket.util"
    if not isinstance(transport_config, str):
        raise ValueError(
            f"'transport_config' must be a string, found {type(transport_config)}"
        )
    # XXX also accept everything Crossbar has in client transport configs? e.g like:
    # { "type": "websocket", "endpoint": {"type": "tcp", "host": "example.com", ...}}
    # XXX what about TLS options? (the above point would address that too)
    if not transport_config.startswith("ws://") and not transport_config.startswith(
        "wss://"
    ):
        raise ValueError("'transport_config' must start with 'ws://' or 'wss://'")
    return None


def check_client_options(options: dict[str, Any]) -> None:
    """
    raises a ValueError if `options` is invalid
    """
    # XXX move me to "autobahn.websocket.util"
    if not isinstance(options, dict):
        raise ValueError("'options' must be a dict")

    # anything that WebSocketClientFactory accepts (at least)
    valid_keys = [
        "origin",
        "protocols",
        "useragent",
        "headers",
        "proxy",
    ]
    for actual_k in options.keys():
        if actual_k not in valid_keys:
            raise ValueError(f"'options' may not contain '{actual_k}'")


def _endpoint_from_config(reactor, factory, transport_config, options):
    # XXX might want some Crossbar code here? e.g. if we allow
    # "transport_config" to be a dict etc.

    # ... passing in the Factory is weird, but that's what parses all
    # the options and the URL currently

    if factory.isSecure:
        # create default client SSL context factory when none given
        from twisted.internet import ssl

        context_factory = ssl.optionsForClientTLS(factory.host)

    if factory.proxy is not None:
        factory.contextFactory = context_factory
        endpoint = endpoints.HostnameEndpoint(
            reactor,
            factory.proxy["host"],
            factory.proxy["port"],
            # timeout,  option?
        )
    else:
        if factory.isSecure:
            from twisted.internet import ssl

            endpoint = endpoints.SSL4ClientEndpoint(
                reactor,
                factory.host,
                factory.port,
                context_factory,
                # timeout,  option?
            )
        else:
            endpoint = endpoints.HostnameEndpoint(  # XXX right? not TCP4ClientEndpoint
                reactor,
                factory.host,
                factory.port,
                # timeout,  option?
                # attemptDelay,  option?
            )
    return endpoint


class _TwistedWebSocketClientAgent(IWebSocketClientAgent):
    """
    This agent creates connections using Twisted
    """

    def __init__(self, reactor):
        self._reactor = reactor

    def open(self, transport_config, options, protocol_class=None):
        """
        Open a new connection.

        :param dict transport_config: valid transport configuration

        :param dict options: additional options for the factory

        :param protocol_class: a callable that returns an instance of
            the protocol (WebSocketClientProtocol if the default None
            is passed in)

        :returns: a Deferred that fires with an instance of
            `protocol_class` (or WebSocketClientProtocol by default)
            that has successfully shaken hands (completed the
            handshake).
        """
        check_transport_config(transport_config)
        check_client_options(options)

        factory = WebSocketClientFactory(
            url=transport_config, reactor=self._reactor, **options
        )
        factory.protocol = (
            WebSocketClientProtocol if protocol_class is None else protocol_class
        )
        # XXX might want "contextFactory" for TLS ...? (or e.g. CA etc options?)

        endpoint = _endpoint_from_config(
            self._reactor, factory, transport_config, options
        )

        rtn_d = Deferred()
        proto_d = endpoint.connect(factory)

        def failed(f):
            rtn_d.errback(f)

        def got_proto(proto):
            def handshake_completed(arg):
                rtn_d.callback(proto)
                return arg

            proto.is_open.addCallbacks(handshake_completed, failed)
            return proto

        proto_d.addCallbacks(got_proto, failed)
        return rtn_d


class WebSocketAdapterProtocol(twisted.internet.protocol.Protocol):
    """
    Adapter class for Twisted WebSocket client and server protocols.

    Called from Twisted:

    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol.connectionMade`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol.connectionLost`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol.dataReceived`

    Called from Network-independent Code (WebSocket implementation):

    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onOpen`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onMessageBegin`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onMessageFrameData`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onMessageFrameEnd`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onMessageEnd`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onMessage`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onPing`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onPong`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._onClose`

    FIXME:

    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._closeConnection`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol._create_transport_details`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol.registerProducer`
    * :meth:`autobahn.twisted.websocket.WebSocketAdapterProtocol.unregisterProducer`
    """

    log = txaio.make_logger()

    peer: str | None = None
    is_server: bool | None = None

    def connectionMade(self) -> None:
        # Twisted networking framework entry point, called by Twisted
        # when the connection is established (either a client or a server)

        # determine preliminary transport details (what is know at this point)
        self._transport_details = create_transport_details(
            self.transport, self.is_server
        )
        self._transport_details.channel_framing = (
            TransportDetails.CHANNEL_FRAMING_WEBSOCKET
        )

        # backward compatibility
        self.peer = self._transport_details.peer

        # try to set "Nagle" option for TCP sockets
        try:
            self.transport.setTcpNoDelay(self.tcpNoDelay)
        except:  # don't touch this! does not work: AttributeError, OSError
            # eg Unix Domain sockets throw Errno 22 on this
            pass

        # ok, now forward to the networking framework independent code for websocket
        self._connectionMade()

        # ok, done!
        self.log.debug(
            '{func} connection established for peer="{peer}"',
            func=hltype(self.connectionMade),
            peer=hlval(self.peer),
        )

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        # Twisted networking framework entry point, called by Twisted
        # when the connection is lost (either a client or a server)

        was_clean = False
        if isinstance(reason.value, ConnectionDone):
            self.log.debug(
                "Connection to/from {peer} was closed cleanly", peer=self.peer
            )
            was_clean = True

        elif _is_tls_error(reason.value):
            self.log.error(_maybe_tls_reason(reason.value))

        elif isinstance(reason.value, ConnectionAborted):
            self.log.debug(
                "Connection to/from {peer} was aborted locally", peer=self.peer
            )

        elif isinstance(reason.value, ConnectionLost):
            message = str(reason.value)
            if hasattr(reason.value, "message"):
                message = reason.value.message
            self.log.debug(
                "Connection to/from {peer} was lost in a non-clean fashion: {message}",
                peer=self.peer,
                message=message,
            )

        # at least: FileDescriptorOverrun, ConnectionFdescWentAway - but maybe others as well?
        else:
            self.log.debug(
                "Connection to/from {peer} lost ({error_type}): {error})",
                peer=self.peer,
                error_type=type(reason.value),
                error=reason.value,
            )

        # ok, now forward to the networking framework independent code for websocket
        self._connectionLost(reason)

        # ok, done!
        if was_clean:
            self.log.debug(
                '{func} connection lost for peer="{peer}", closed cleanly',
                func=hltype(self.connectionLost),
                peer=hlval(self.peer),
            )
        else:
            self.log.debug(
                '{func} connection lost for peer="{peer}", closed with error {reason}',
                func=hltype(self.connectionLost),
                peer=hlval(self.peer),
                reason=reason,
            )

    def dataReceived(self, data: bytes) -> None:
        self.log.debug(
            '{func} received {data_len} bytes for peer="{peer}"',
            func=hltype(self.dataReceived),
            peer=hlval(self.peer),
            data_len=hlval(len(data)),
        )

        # bytes received from Twisted, forward to the networking framework independent code for websocket
        self._dataReceived(data)

    def _closeConnection(self, abort: bool=False) -> None:
        if abort and hasattr(self.transport, "abortConnection"):
            self.transport.abortConnection()
        else:
            # e.g. ProcessProtocol lacks abortConnection()
            self.transport.loseConnection()

    def _onOpen(self) -> None:
        if self._transport_details.is_secure:
            # now that the TLS opening handshake is complete, the actual TLS channel ID
            # will be available. make sure to set it!
            channel_id = {
                "tls-unique": transport_channel_id(
                    self.transport, self._transport_details.is_server, "tls-unique"
                ),
            }
            self._transport_details.channel_id = channel_id

        self.onOpen()

    def _onMessageBegin(self, isBinary: bool) -> None:
        self.onMessageBegin(isBinary)

    def _onMessageFrameBegin(self, length: int) -> None:
        self.onMessageFrameBegin(length)

    def _onMessageFrameData(self, payload) -> None:
        self.onMessageFrameData(payload)

    def _onMessageFrameEnd(self) -> None:
        self.onMessageFrameEnd()

    def _onMessageFrame(self, payload) -> None:
        self.onMessageFrame(payload)

    def _onMessageEnd(self) -> None:
        self.onMessageEnd()

    def _onMessage(self, payload, isBinary: bool) -> None:
        self.onMessage(payload, isBinary)

    def _onPing(self, payload) -> None:
        self.onPing(payload)

    def _onPong(self, payload) -> None:
        self.onPong(payload)

    def _onClose(self, wasClean: bool, code, reason) -> None:
        self.onClose(wasClean, code, reason)

    def registerProducer(self, producer, streaming) -> None:
        """
        Register a Twisted producer with this protocol.

        :param producer: A Twisted push or pull producer.
        :type producer: object
        :param streaming: Producer type.
        :type streaming: bool
        """
        self.transport.registerProducer(producer, streaming)

    def unregisterProducer(self) -> None:
        """
        Unregister Twisted producer with this protocol.
        """
        self.transport.unregisterProducer()


@public
class WebSocketServerProtocol(
    WebSocketAdapterProtocol, protocol.WebSocketServerProtocol
):
    """
    Base class for Twisted-based WebSocket server protocols.

    Implements :class:`autobahn.websocket.interfaces.IWebSocketChannel`.
    """

    log = txaio.make_logger()
    is_server = True

    # def onConnect(self, request: ConnectionRequest) -> Union[Optional[str], Tuple[Optional[str], Dict[str, str]]]:
    #     pass


@public
class WebSocketClientProtocol(
    WebSocketAdapterProtocol, protocol.WebSocketClientProtocol
):
    """
    Base class for Twisted-based WebSocket client protocols.

    Implements :class:`autobahn.websocket.interfaces.IWebSocketChannel`.
    """

    log = txaio.make_logger()
    is_server = False

    def _onConnect(self, response: ConnectionResponse):
        self.log.debug(
            "{meth}(response={response})",
            meth=hltype(self._onConnect),
            response=response,
        )
        return self.onConnect(response)

    def startTLS(self):
        self.log.debug("Starting TLS upgrade")
        self.transport.startTLS(self.factory.contextFactory)


class WebSocketAdapterFactory:
    """
    Adapter class for Twisted-based WebSocket client and server factories.
    """


@public
class WebSocketServerFactory(
    WebSocketAdapterFactory,
    protocol.WebSocketServerFactory,
    twisted.internet.protocol.ServerFactory,
):
    """
    Base class for Twisted-based WebSocket server factories.

    Implements :class:`autobahn.websocket.interfaces.IWebSocketServerChannelFactory`
    """

    log = txaio.make_logger()

    def __init__(self, *args, **kwargs):
        """

        .. note::
            In addition to all arguments to the constructor of
            :meth:`autobahn.websocket.interfaces.IWebSocketServerChannelFactory`,
            you can supply a ``reactor`` keyword argument to specify the
            Twisted reactor to be used.
        """
        # lazy import to avoid reactor install upon module import
        reactor = kwargs.pop("reactor", None)
        if reactor is None:
            from twisted.internet import reactor
        self.reactor = reactor

        protocol.WebSocketServerFactory.__init__(self, *args, **kwargs)


@public
class WebSocketClientFactory(
    WebSocketAdapterFactory,
    protocol.WebSocketClientFactory,
    twisted.internet.protocol.ClientFactory,
):
    """
    Base class for Twisted-based WebSocket client factories.

    Implements :class:`autobahn.websocket.interfaces.IWebSocketClientChannelFactory`
    """

    log = txaio.make_logger()

    def __init__(self, *args, **kwargs):
        """

        .. note::
            In addition to all arguments to the constructor of
            :func:`autobahn.websocket.interfaces.IWebSocketClientChannelFactory`,
            you can supply a ``reactor`` keyword argument to specify the
            Twisted reactor to be used.
        """
        # lazy import to avoid reactor install upon module import
        reactor = kwargs.pop("reactor", None)
        if reactor is None:
            from twisted.internet import reactor
        self.reactor = reactor

        protocol.WebSocketClientFactory.__init__(self, *args, **kwargs)
        # we must up-call *before* we set up the contextFactory
        # because we need self.host etc to be set properly.
        if self.isSecure and self.proxy is not None:
            # if we have a proxy, then our factory will be used to
            # create the connection after CONNECT and if it's doing
            # TLS it needs a contextFactory
            from twisted.internet import ssl

            self.contextFactory = ssl.optionsForClientTLS(self.host)
        # NOTE: there's thus no way to send in our own
        # context-factory, nor any TLS options.

        # Possibly we should allow 'proxy' to contain an actual
        # IStreamClientEndpoint instance instead of configuration for
        # how to make one


@implementer(ITransport)
class WrappingWebSocketAdapter:
    """
    An adapter for stream-based transport over WebSocket.

    This follows `websockify <https://github.com/kanaka/websockify>`_
    and should be compatible with that.

    It uses WebSocket subprotocol negotiation and supports the
    following WebSocket subprotocols:

      - ``binary`` (or a compatible subprotocol)
      - ``base64``

    Octets are either transmitted as the payload of WebSocket binary
    messages when using the ``binary`` subprotocol (or an alternative
    binary compatible subprotocol), or encoded with Base64 and then
    transmitted as the payload of WebSocket text messages when using
    the ``base64`` subprotocol.
    """

    def onConnect(self, requestOrResponse):
        # Negotiate either the 'binary' or the 'base64' WebSocket subprotocol
        if isinstance(requestOrResponse, ConnectionRequest):
            request = requestOrResponse
            for p in request.protocols:
                if p in self.factory._subprotocols:
                    self._binaryMode = p != "base64"
                    return p
            raise ConnectionDeny(
                ConnectionDeny.NOT_ACCEPTABLE,
                f"this server only speaks {self.factory._subprotocols} WebSocket subprotocols",
            )
        elif isinstance(requestOrResponse, ConnectionResponse):
            response = requestOrResponse
            if response.protocol not in self.factory._subprotocols:
                self._fail_connection(
                    protocol.WebSocketProtocol.CLOSE_STATUS_CODE_PROTOCOL_ERROR,
                    f"this client only speaks {self.factory._subprotocols} WebSocket subprotocols",
                )
            self._binaryMode = response.protocol != "base64"
        else:
            # should not arrive here
            raise Exception("logic error")

    def onOpen(self) -> None:
        self._proto.connectionMade()

    def onMessage(self, payload: bytes, isBinary: bool) -> None:
        if isBinary != self._binaryMode:
            self._fail_connection(
                protocol.WebSocketProtocol.CLOSE_STATUS_CODE_UNSUPPORTED_DATA,
                "message payload type does not match the negotiated subprotocol",
            )
        else:
            if not isBinary:
                try:
                    payload = b64decode(payload)
                except Exception as e:
                    self._fail_connection(
                        protocol.WebSocketProtocol.CLOSE_STATUS_CODE_INVALID_PAYLOAD,
                        f"message payload base64 decoding error: {e}",
                    )
            self._proto.dataReceived(payload)

    # noinspection PyUnusedLocal
    def onClose(self, wasClean, code, reason):
        self._proto.connectionLost(None)

    def write(self, data: bytes) -> None:
        # part of ITransport
        assert type(data) == bytes
        if self._binaryMode:
            self.sendMessage(data, isBinary=True)
        else:
            data = b64encode(data)
            self.sendMessage(data, isBinary=False)

    def writeSequence(self, data: bytes) -> None:
        # part of ITransport
        for d in data:
            self.write(d)

    def loseConnection(self) -> None:
        # part of ITransport
        self.sendClose()

    def getPeer(self):
        # part of ITransport
        return self.transport.getPeer()

    def getHost(self):
        # part of ITransport
        return self.transport.getHost()


class WrappingWebSocketServerProtocol(
    WrappingWebSocketAdapter, WebSocketServerProtocol
):
    """
    Server protocol for stream-based transport over WebSocket.
    """


class WrappingWebSocketClientProtocol(
    WrappingWebSocketAdapter, WebSocketClientProtocol
):
    """
    Client protocol for stream-based transport over WebSocket.
    """


class WrappingWebSocketServerFactory(WebSocketServerFactory):
    """
    Wrapping server factory for stream-based transport over WebSocket.
    """

    def __init__(
        self,
        factory,
        url,
        reactor=None,
        enableCompression=True,
        autoFragmentSize=0,
        subprotocol=None,
    ):
        """

        :param factory: Stream-based factory to be wrapped.
        :type factory: A subclass of ``twisted.internet.protocol.Factory``
        :param url: WebSocket URL of the server this server factory will work for.
        :type url: unicode
        """
        self._factory = factory
        self._subprotocols = ["binary", "base64"]
        if subprotocol:
            self._subprotocols.append(subprotocol)

        WebSocketServerFactory.__init__(
            self, url=url, reactor=reactor, protocols=self._subprotocols
        )

        # automatically fragment outgoing traffic into WebSocket frames
        # of this size
        self.setProtocolOptions(autoFragmentSize=autoFragmentSize)

        # play nice and perform WS closing handshake
        self.setProtocolOptions(failByDrop=False)

        if enableCompression:
            # Enable WebSocket extension "permessage-deflate".

            # Function to accept offers from the client ..
            def accept(offers):
                for offer in offers:
                    if isinstance(offer, PerMessageDeflateOffer):
                        return PerMessageDeflateOfferAccept(offer)

            self.setProtocolOptions(perMessageCompressionAccept=accept)

    def buildProtocol(self, addr):
        proto = WrappingWebSocketServerProtocol()
        proto.factory = self
        proto._proto = self._factory.buildProtocol(addr)
        proto._proto.transport = proto
        return proto

    def startFactory(self):
        self._factory.startFactory()
        WebSocketServerFactory.startFactory(self)

    def stopFactory(self):
        self._factory.stopFactory()
        WebSocketServerFactory.stopFactory(self)


class WrappingWebSocketClientFactory(WebSocketClientFactory):
    """
    Wrapping client factory for stream-based transport over WebSocket.
    """

    def __init__(
        self,
        factory,
        url,
        reactor=None,
        enableCompression=True,
        autoFragmentSize=0,
        subprotocol=None,
    ):
        """

        :param factory: Stream-based factory to be wrapped.
        :type factory: A subclass of ``twisted.internet.protocol.Factory``
        :param url: WebSocket URL of the server this client factory will connect to.
        :type url: unicode
        """
        self._factory = factory
        self._subprotocols = ["binary", "base64"]
        if subprotocol:
            self._subprotocols.append(subprotocol)

        WebSocketClientFactory.__init__(
            self, url=url, reactor=reactor, protocols=self._subprotocols
        )

        # automatically fragment outgoing traffic into WebSocket frames
        # of this size
        self.setProtocolOptions(autoFragmentSize=autoFragmentSize)

        # play nice and perform WS closing handshake
        self.setProtocolOptions(failByDrop=False)

        if enableCompression:
            # Enable WebSocket extension "permessage-deflate".

            # The extensions offered to the server ..
            offers = [PerMessageDeflateOffer()]
            self.setProtocolOptions(perMessageCompressionOffers=offers)

            # Function to accept responses from the server ..
            def accept(response):
                if isinstance(response, PerMessageDeflateResponse):
                    return PerMessageDeflateResponseAccept(response)

            self.setProtocolOptions(perMessageCompressionAccept=accept)

    def buildProtocol(self, addr):
        proto = WrappingWebSocketClientProtocol()
        proto.factory = self
        proto._proto = self._factory.buildProtocol(addr)
        proto._proto.transport = proto
        return proto


@public
def connectWS(factory, contextFactory=None, timeout=30, bindAddress=None):
    """
    Establish WebSocket connection to a server. The connection parameters like target
    host, port, resource and others are provided via the factory.

    :param factory: The WebSocket protocol factory to be used for creating client protocol instances.
    :type factory: An :class:`autobahn.websocket.WebSocketClientFactory` instance.

    :param contextFactory: SSL context factory, required for secure WebSocket connections ("wss").
    :type contextFactory: A `twisted.internet.ssl.ClientContextFactory <http://twistedmatrix.com/documents/current/api/twisted.internet.ssl.ClientContextFactory.html>`_ instance.

    :param timeout: Number of seconds to wait before assuming the connection has failed.
    :type timeout: int

    :param bindAddress: A (host, port) tuple of local address to bind to, or None.
    :type bindAddress: tuple

    :returns: The connector.
    :rtype: An object which implements `twisted.interface.IConnector <http://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IConnector.html>`_.
    """
    # lazy import to avoid reactor install upon module import
    if hasattr(factory, "reactor"):
        reactor = factory.reactor
    else:
        from twisted.internet import reactor

    if factory.isSecure:
        if contextFactory is None:
            # create default client SSL context factory when none given
            from twisted.internet import ssl

            contextFactory = ssl.ClientContextFactory()

    if factory.proxy is not None:
        factory.contextFactory = contextFactory
        conn = reactor.connectTCP(
            factory.proxy["host"], factory.proxy["port"], factory, timeout, bindAddress
        )
    else:
        if factory.isSecure:
            conn = reactor.connectSSL(
                factory.host,
                factory.port,
                factory,
                contextFactory,
                timeout,
                bindAddress,
            )
        else:
            conn = reactor.connectTCP(
                factory.host, factory.port, factory, timeout, bindAddress
            )
    return conn


@public
def listenWS(factory, contextFactory=None, backlog=50, interface=""):
    """
    Listen for incoming WebSocket connections from clients. The connection parameters like
    listening port and others are provided via the factory.

    :param factory: The WebSocket protocol factory to be used for creating server protocol instances.
    :type factory: An :class:`autobahn.websocket.WebSocketServerFactory` instance.

    :param contextFactory: SSL context factory, required for secure WebSocket connections ("wss").
    :type contextFactory: A twisted.internet.ssl.ContextFactory.

    :param backlog: Size of the listen queue.
    :type backlog: int

    :param interface: The interface (derived from hostname given) to bind to, defaults to '' (all).
    :type interface: str

    :returns: The listening port.
    :rtype: An object that implements `twisted.interface.IListeningPort <http://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IListeningPort.html>`_.
    """
    # lazy import to avoid reactor install upon module import
    if hasattr(factory, "reactor"):
        reactor = factory.reactor
    else:
        from twisted.internet import reactor

    if factory.isSecure:
        if contextFactory is None:
            raise Exception(
                "Secure WebSocket listen requested, but no SSL context factory given"
            )
        listener = reactor.listenSSL(
            factory.port, factory, contextFactory, backlog, interface
        )
    else:
        listener = reactor.listenTCP(factory.port, factory, backlog, interface)
    return listener


@public
class WampWebSocketServerProtocol(
    websocket.WampWebSocketServerProtocol, WebSocketServerProtocol
):
    """
    Twisted-based WAMP-over-WebSocket server protocol.

    Implements:

    * :class:`autobahn.wamp.interfaces.ITransport`
    """


@public
class WampWebSocketServerFactory(
    websocket.WampWebSocketServerFactory, WebSocketServerFactory
):
    """
    Twisted-based WAMP-over-WebSocket server protocol factory.
    """

    protocol = WampWebSocketServerProtocol

    def __init__(self, factory, *args, **kwargs):
        """

        :param factory: A callable that produces instances that implement
            :class:`autobahn.wamp.interfaces.ITransportHandler`
        :type factory: callable

        :param serializers: A list of WAMP serializers to use (or ``None``
            for all available serializers).
        :type serializers: list of objects implementing
            :class:`autobahn.wamp.interfaces.ISerializer`
        """

        serializers = kwargs.pop("serializers", None)

        websocket.WampWebSocketServerFactory.__init__(self, factory, serializers)

        kwargs["protocols"] = self._protocols

        # noinspection PyCallByClass
        WebSocketServerFactory.__init__(self, *args, **kwargs)


@public
class WampWebSocketClientProtocol(
    websocket.WampWebSocketClientProtocol, WebSocketClientProtocol
):
    """
    Twisted-based WAMP-over-WebSocket client protocol.

    Implements:

    * :class:`autobahn.wamp.interfaces.ITransport`
    """


@public
class WampWebSocketClientFactory(
    websocket.WampWebSocketClientFactory, WebSocketClientFactory
):
    """
    Twisted-based WAMP-over-WebSocket client protocol factory.
    """

    protocol = WampWebSocketClientProtocol

    def __init__(self, factory, *args, **kwargs):
        """

        :param factory: A callable that produces instances that implement
            :class:`autobahn.wamp.interfaces.ITransportHandler`
        :type factory: callable

        :param serializer: The WAMP serializer to use (or ``None`` for
           "best" serializer, chosen as the first serializer available from
           this list: CBOR, MessagePack, UBJSON, JSON).
        :type serializer: object implementing :class:`autobahn.wamp.interfaces.ISerializer`
        """

        serializers = kwargs.pop("serializers", None)

        websocket.WampWebSocketClientFactory.__init__(self, factory, serializers)

        kwargs["protocols"] = self._protocols

        WebSocketClientFactory.__init__(self, *args, **kwargs)

        # Reduce the factory logs noise
        self.noisy = False



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import util
from ._version import __version__
from .builder import Builder
from .compat import range_func as compat_range
from .table import Table

# --- Autobahn additions for bundled flatc support ---
import re
from ._git_version import __git_version__


def version() -> tuple[int, int, int, int | None, str | None]:
    """
    Return the version of the vendored FlatBuffers runtime as a 5-tuple
    ``(major, minor, patch, commits, commit_hash)``.

    The primary source is ``__git_version__`` (``git describe`` output), which on a
    genuine dev/git build carries rich detail:

    1. "v25.9.23"              -> (25, 9, 23, None, None)       # Release (Named Tag, CalVer Year.Month.Day)
    2. "v25.9.23-71"           -> (25, 9, 23, 71, None)         # 71 commits ahead of the Release v25.9.23
    3. "v25.9.23-71-g19b2300f" -> (25, 9, 23, 71, "19b2300f")   # dito, with Git commit hash

    On installed wheels / shallow clones ``__git_version__`` may be a bare commit hash
    or "unknown" (``git describe`` found no tags); in that case fall back to the always
    reliable static ``__version__`` and return ``(major, minor, patch, None, None)``.
    """
    pattern = r"^v(\d+)\.(\d+)\.(\d+)(?:-(\d+))?(?:-g([0-9a-f]+))?$"
    match = re.match(pattern, __git_version__)
    if match:
        major, minor, patch, commits, commit_hash = match.groups()
        commits_int = int(commits) if commits else None
        return (int(major), int(minor), int(patch), commits_int, commit_hash)
    # Fallback: static vendored __version__ ("YY.MM.DD", no 'v' prefix) is reliable on
    # wheels where __git_version__ may be a bare hash / "unknown". No end-anchor so a
    # trailing dev suffix is tolerated.
    fallback = re.match(r"^v?(\d+)\.(\d+)\.(\d+)", __version__)
    if fallback:
        major, minor, patch = fallback.groups()
        return (int(major), int(minor), int(patch), None, None)
    return (0, 0, 0, None, None)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# TLS channel binding type (see: RFC5929 https://www.rfc-editor.org/rfc/rfc5929 and RFC9266 https://www.rfc-editor.org/rfc/rfc9266).
class TLSChannelBinding(object):
    # Not set / no channel binding.
    NULL = 0
    TLS_UNIQUE = 1
    # RFC 5929 `tls-unique-for-telnet`. Only relevant to TELNET / TELNET AUTH usage. See RFC 5929 for details and applicability guidance.
    TLS_UNIQUE_TELNET = 2
    # RFC 5929 `tls-server-end-point`. Only available when a server certificate is used (i.e., cipher suites that include the Certificate handshake message / PKIX). Not applicable to OpenPGP server certificates. Recommended for situations where server-side proxies must interoperate without changes.
    TLS_SERVER_ENDPOINT = 3
    # RFC 9266 `tls-exporter`. Defined to address TLS 1.3 (where the older tls-unique semantics are not reliably available). RFC 9266 updates defaults for TLS ≥ 1.3: when channel bindings are used for TLS 1.3, tls-exporter is the mandatory/expected mechanism. Not defined for connections where TLS renegotiation is enabled.
    TLS_EXPORTER = 4



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP Payload Passthru (PPT) scheme (renamed from `Payload`).
class PPTScheme(object):
    # Unset (plain WAMP application payload)
    NONE = 0
    # WAMP-cryptobox end-to-end encrypted application payload
    CRYPTOBOX = 1
    # MQTT passthrough-mode application payload
    MQTT = 2
    # XBR end-to-end encrypted and Ethereum anchored application payload
    XBR = 3
    # Raw pass-through of app payload, uninterpreted in any way.
    OPAQUE = 4



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import binascii
import json
import platform
import re
import unittest
from unittest.mock import Mock

from autobahn.wamp import auth, types

# these test vectors are all for HMAC-SHA1
PBKDF2_TEST_VECTORS = [
    # From RFC 6070
    (b"password", b"salt", 1, 20, "0c60c80f961f0e71f3a9b524af6012062fe037a6"),
    (b"password", b"salt", 2, 20, "ea6c014dc72d6f8ccd1ed92ace1d41f0d8de8957"),
    # From Crypt-PBKDF2
    (b"password", b"ATHENA.MIT.EDUraeburn", 1, 16, "cdedb5281bb2f801565a1122b2563515"),
    (
        b"password",
        b"ATHENA.MIT.EDUraeburn",
        1,
        32,
        "cdedb5281bb2f801565a1122b25635150ad1f7a04bb9f3a333ecc0e2e1f70837",
    ),
    (b"password", b"ATHENA.MIT.EDUraeburn", 2, 16, "01dbee7f4a9e243e988b62c73cda935d"),
    (
        b"password",
        b"ATHENA.MIT.EDUraeburn",
        2,
        32,
        "01dbee7f4a9e243e988b62c73cda935da05378b93244ec8f48a99e61ad799d86",
    ),
    (
        b"password",
        b"ATHENA.MIT.EDUraeburn",
        1200,
        32,
        "5c08eb61fdf71e4e4ec3cf6ba1f5512ba7e52ddbc5e5142f708a31e2e62b1e13",
    ),
    (
        b"X" * 64,
        b"pass phrase equals block size",
        1200,
        32,
        "139c30c0966bc32ba55fdbf212530ac9c5ec59f1a452f5cc9ad940fea0598ed1",
    ),
    (
        b"X" * 65,
        b"pass phrase exceeds block size",
        1200,
        32,
        "9ccad6d468770cd51b10e6a68721be611a8b4d282601db3b36be9246915ec82a",
    ),
]

if platform.python_implementation() != "PyPy":
    # the following fails on PyPy: "RuntimeError: maximum recursion depth exceeded"
    PBKDF2_TEST_VECTORS.extend(
        [
            # From RFC 6070
            (
                b"password",
                b"salt",
                4096,
                20,
                "4b007901b765489abead49d926f721d065a429c1",
            ),
            (
                b"passwordPASSWORDpassword",
                b"saltSALTsaltSALTsaltSALTsaltSALTsalt",
                4096,
                25,
                "3d2eec4fe41c849b80c8d83662c0e44a8b291a964cf2f07038",
            ),
            (
                b"pass\x00word",
                b"sa\x00lt",
                4096,
                16,
                "56fa6aa75548099dcc37d7f03425e0c3",
            ),
            # This one is from the RFC but it just takes for ages
            # (b'password', b'salt', 16777216, 20, 'eefe3d61cd4da4e4e9945b3d6ba2158c2634e984'),
        ]
    )


class TestWampAuthHelpers(unittest.TestCase):
    def test_pbkdf2(self):
        for tv in PBKDF2_TEST_VECTORS:
            result = auth.pbkdf2(tv[0], tv[1], tv[2], tv[3], "sha1")
            self.assertEqual(type(result), bytes)
            self.assertEqual(binascii.hexlify(result).decode("ascii"), tv[4])

    def test_generate_totp_secret_default(self):
        secret = auth.generate_totp_secret()
        self.assertEqual(type(secret), str)
        self.assertEqual(len(secret), 10 * 8 / 5)

    def test_generate_totp_secret_length(self):
        for length in [5, 10, 20, 30, 40, 50]:
            secret = auth.generate_totp_secret(length)
            self.assertEqual(type(secret), str)
            self.assertEqual(len(secret), length * 8 / 5)

    def test_compute_totp(self):
        pat = re.compile(r"\d\d\d\d\d\d")
        secret = "MFRGGZDFMZTWQ2LK"
        signature = auth.compute_totp(secret)
        self.assertEqual(type(signature), str)
        self.assertTrue(pat.match(signature) is not None)

    def test_compute_totp_offset(self):
        pat = re.compile(r"\d\d\d\d\d\d")
        secret = "MFRGGZDFMZTWQ2LK"
        for offset in range(-10, 10):
            signature = auth.compute_totp(secret, offset)
            self.assertEqual(type(signature), str)
            self.assertTrue(pat.match(signature) is not None)

    def test_derive_key(self):
        secret = "L3L1YUE8Txlw"
        salt = "salt123"
        key = auth.derive_key(secret.encode("utf8"), salt.encode("utf8"))
        self.assertEqual(type(key), bytes)
        self.assertEqual(key, b"qzcdsr9uu/L5hnss3kjNTRe490ETgA70ZBaB5rvnJ5Y=")

    def test_generate_wcs_default(self):
        secret = auth.generate_wcs()
        self.assertEqual(type(secret), bytes)
        self.assertEqual(len(secret), 14)

    def test_generate_wcs_length(self):
        for length in [5, 10, 20, 30, 40, 50]:
            secret = auth.generate_wcs(length)
            self.assertEqual(type(secret), bytes)
            self.assertEqual(len(secret), length)

    def test_compute_wcs(self):
        secret = "L3L1YUE8Txlw"
        challenge = json.dumps([1, 2, 3], ensure_ascii=False).encode("utf8")
        signature = auth.compute_wcs(secret.encode("utf8"), challenge)
        self.assertEqual(type(signature), bytes)
        self.assertEqual(signature, b"1njQtmmeYO41N5EWEzD2kAjjEKRZ5kPZt/TzpYXOzR0=")


@unittest.skipIf(not auth.HAS_ARGON, "no Argon2 library")
class TestScram(unittest.TestCase):
    def test_argon2id_static(self):
        # re-generate from the official argon2 tools:
        # echo -n "p4ssw0rd" | argon2 '1234567890abcdef' -id -t 32 -m 9 -p 1 -l 32
        expected = binascii.unhexlify(
            "ee4a8acf9d5958354fb79a95ae20692d05e42591ba49fae85eb6700e8b0ed293"
        )
        raw_hash = auth._hash_argon2id13_secret(
            b"p4ssw0rd",
            binascii.b2a_base64(b"1234567890abcdef"),  # ours takes base64-encoded salt
            32,  # this is WAY TOO SMALL; for production, use 4096 or higher
            512,  # note that the argon2 utility takes a "power of 2", so "-m 9" above == 512
        )
        decoded_hash = binascii.a2b_base64(raw_hash + b"==\n")
        self.assertEqual(expected, decoded_hash)

    def test_pbkdf2_static(self):
        expected = binascii.unhexlify(
            "f6991a28c75f43751e0d75499fd7b8649f659118ddc1d61cee5883af547d15f5"
        )
        # 8 iterations is WAY TOO FEW for production; this is a test
        raw_hash = auth._hash_pbkdf2_secret(b"p4ssw0rd", b"1234567890abcdef", 8)
        self.assertEqual(raw_hash, expected)

    def test_basic(self):
        scram = auth.AuthScram(
            nonce="1234567890abcdef",
            kdf="argon2id13",
            salt=binascii.b2a_hex(b"1234567890abcdef").decode("ascii"),
            iterations=32,  # far too few; use 4096 or more for production
            memory=512,
            password="p4ssw0rd",
            authid="username",
        )
        # thought: if we could import crossbar code here, we could
        # test the "other side" of this with fewer mocks
        # (i.e. hard-coding the client nonce)
        scram._client_nonce = binascii.b2a_hex(b"1234567890abcdef").decode("ascii")
        self.assertEqual(
            {"nonce": "31323334353637383930616263646566"},
            scram.authextra,
        )

        challenge = types.Challenge(
            "scram",
            {
                "nonce": "1234567890abcdeffedcba0987654321",
                "kdf": "argon2id-13",
                "salt": binascii.b2a_hex(b"1234567890abcdef").decode("ascii"),
                "iterations": 32,
                "memory": 512,
            },
        )
        reply = scram.on_challenge(Mock(), challenge)
        self.assertEqual(
            b"f5r3loERzGVSuimE+lvO0bWna2zyswBo0HrZkaaEy38=",
            reply,
        )

        authextra = dict(
            scram_server_signature=b"f5r3loERzGVSuimE+lvO0bWna2zyswBo0HrZkaaEy38=",
        )
        scram.on_welcome(Mock(), authextra)

    def test_no_memory_arg(self):
        scram = auth.AuthScram(
            nonce="1234567890abcdef",
            kdf="argon2id13",
            salt=binascii.b2a_hex(b"1234567890abcdef").decode("ascii"),
            iterations=4096,
            memory=512,
            password="p4ssw0rd",
            authid="username",
        )
        scram.authextra

        with self.assertRaises(ValueError) as ctx:
            challenge = types.Challenge(
                "scram",
                {
                    "nonce": "1234567890abcdeffedcba0987654321",
                    "kdf": "argon2id-13",
                    "salt": binascii.b2a_hex(b"1234567890abcdef"),
                    "iterations": 4096,
                    # no 'memory' key
                },
            )
            scram.on_challenge(Mock(), challenge)
        self.assertIn("requires 'memory' parameter", str(ctx.exception))

    def test_unknown_arg(self):
        scram = auth.AuthScram(
            nonce="1234567890abcdef",
            kdf="argon2id13",
            salt=binascii.b2a_hex(b"1234567890abcdef"),
            iterations=4096,
            memory=512,
            password="p4ssw0rd",
            authid="username",
        )
        scram.authextra

        with self.assertRaises(RuntimeError) as ctx:
            challenge = types.Challenge(
                "scram",
                {
                    "nonce": "1234567890abcdeffedcba0987654321",
                    "kdf": "argon2id-13",
                    "salt": binascii.b2a_hex(b"1234567890abcdef"),
                    "iterations": 4096,
                    "memory": 512,
                    "an_invalid_key": None,
                },
            )
            scram.on_challenge(Mock(), challenge)
        self.assertIn("an_invalid_key", str(ctx.exception))



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import bz2

from autobahn.exception import PayloadExceededError
from autobahn.websocket.compress_base import (
    PerMessageCompress,
    PerMessageCompressOffer,
    PerMessageCompressOfferAccept,
    PerMessageCompressResponse,
    PerMessageCompressResponseAccept,
)

__all__ = (
    "PerMessageBzip2",
    "PerMessageBzip2Mixin",
    "PerMessageBzip2Offer",
    "PerMessageBzip2OfferAccept",
    "PerMessageBzip2Response",
    "PerMessageBzip2ResponseAccept",
)


class PerMessageBzip2Mixin:
    """
    Mixin class for this extension.
    """

    EXTENSION_NAME = "permessage-bzip2"
    """
    Name of this WebSocket extension.
    """

    COMPRESS_LEVEL_PERMISSIBLE_VALUES = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    Permissible value for compression level parameter.
    """


class PerMessageBzip2Offer(PerMessageCompressOffer, PerMessageBzip2Mixin):
    """
    Set of extension parameters for `permessage-bzip2` WebSocket extension
    offered by a client to a server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension offer for `permessage-bzip2` provided by a client to a server.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: object -- A new instance of :class:`autobahn.compress.PerMessageBzip2Offer`.
        """
        # extension parameter defaults
        accept_max_compress_level = False
        request_max_compress_level = 0

        # verify/parse client ("client-to-server direction") parameters of permessage-bzip2 offer
        for p in params:
            if len(params[p]) > 1:
                raise Exception(
                    f"multiple occurrence of extension parameter '{p}' for extension '{cls.EXTENSION_NAME}'"
                )

            val = params[p][0]

            if p == "client_max_compress_level":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        f"illegal extension parameter value '{val}' for parameter '{p}' of extension '{cls.EXTENSION_NAME}'"
                    )
                else:
                    accept_max_compress_level = True

            elif p == "server_max_compress_level":
                try:
                    val = int(val)
                except:
                    raise Exception(
                        f"illegal extension parameter value '{val}' for parameter '{p}' of extension '{cls.EXTENSION_NAME}'"
                    )
                if val not in PerMessageBzip2Mixin.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                    raise Exception(
                        f"illegal extension parameter value '{val}' for parameter '{p}' of extension '{cls.EXTENSION_NAME}'"
                    )
                else:
                    request_max_compress_level = val

            else:
                raise Exception(
                    f"illegal extension parameter '{p}' for extension '{cls.EXTENSION_NAME}'"
                )

        offer = cls(accept_max_compress_level, request_max_compress_level)
        return offer

    def __init__(self, accept_max_compress_level=True, request_max_compress_level=0):
        """
        Constructor.

        :param accept_max_compress_level: Iff true, client accepts "maximum compression level" parameter.
        :type accept_max_compress_level: bool
        :param request_max_compress_level: Iff non-zero, client requests given "maximum compression level" - must be 1-9.
        :type request_max_compress_level: int
        """
        if type(accept_max_compress_level) != bool:
            raise Exception(
                f"invalid type {type(accept_max_compress_level)} for accept_max_compress_level"
            )

        self.accept_max_compress_level = accept_max_compress_level

        if (
            request_max_compress_level != 0
            and request_max_compress_level not in self.COMPRESS_LEVEL_PERMISSIBLE_VALUES
        ):
            raise Exception(
                f"invalid value {request_max_compress_level} for request_max_compress_level - permissible values {self.COMPRESS_LEVEL_PERMISSIBLE_VALUES}"
            )

        self.request_max_compress_level = request_max_compress_level

    def get_extension_string(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: PMCE configuration string.
        :rtype: str
        """
        pmce_string = self.EXTENSION_NAME
        if self.accept_max_compress_level:
            pmce_string += "; client_max_compress_level"
        if self.request_max_compress_level != 0:
            pmce_string += (
                f"; server_max_compress_level={self.request_max_compress_level}"
            )
        return pmce_string

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "accept_max_compress_level": self.accept_max_compress_level,
            "request_max_compress_level": self.request_max_compress_level,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            f"PerMessageBzip2Offer(accept_max_compress_level = {self.accept_max_compress_level}, request_max_compress_level = {self.request_max_compress_level})"
        )


class PerMessageBzip2OfferAccept(PerMessageCompressOfferAccept, PerMessageBzip2Mixin):
    """
    Set of parameters with which to accept an `permessage-bzip2` offer
    from a client by a server.
    """

    def __init__(self, offer, request_max_compress_level=0, compress_level=None):
        """
        Constructor.

        :param offer: The offer being accepted.
        :type offer: Instance of :class:`autobahn.compress.PerMessageBzip2Offer`.
        :param request_max_compress_level: Iff non-zero, server requests given "maximum compression level" - must be 1-9.
        :type request_max_compress_level: int
        :param compress_level: Override server ("server-to-client direction") compress level (this must be compatible with offer).
        :type compress_level: int
        """
        if not isinstance(offer, PerMessageBzip2Offer):
            raise Exception(f"invalid type {type(offer)} for offer")

        self.offer = offer

        if (
            request_max_compress_level != 0
            and request_max_compress_level not in self.COMPRESS_LEVEL_PERMISSIBLE_VALUES
        ):
            raise Exception(
                f"invalid value {request_max_compress_level} for request_max_compress_level - permissible values {self.COMPRESS_LEVEL_PERMISSIBLE_VALUES}"
            )

        if request_max_compress_level != 0 and not offer.accept_max_compress_level:
            raise Exception(
                f"invalid value {request_max_compress_level} for request_max_compress_level - feature unsupported by client"
            )

        self.request_max_compress_level = request_max_compress_level

        if compress_level is not None:
            if compress_level not in self.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                raise Exception(
                    f"invalid value {compress_level} for compress_level - permissible values {self.COMPRESS_LEVEL_PERMISSIBLE_VALUES}"
                )

            if (
                offer.request_max_compress_level != 0
                and compress_level > offer.request_max_compress_level
            ):
                raise Exception(
                    f"invalid value {compress_level} for compress_level - client requested lower maximum value"
                )

        self.compress_level = compress_level

    def get_extension_string(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: PMCE configuration string.
        :rtype: str
        """
        pmce_string = self.EXTENSION_NAME
        if self.offer.request_max_compress_level != 0:
            pmce_string += (
                f"; server_max_compress_level={self.offer.request_max_compress_level}"
            )
        if self.request_max_compress_level != 0:
            pmce_string += (
                f"; client_max_compress_level={self.request_max_compress_level}"
            )
        return pmce_string

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "offer": self.offer.__json__(),
            "request_max_compress_level": self.request_max_compress_level,
            "compress_level": self.compress_level,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            f"PerMessageBzip2Accept(offer = {self.offer.__repr__()}, request_max_compress_level = {self.request_max_compress_level}, compress_level = {self.compress_level})"
        )


class PerMessageBzip2Response(PerMessageCompressResponse, PerMessageBzip2Mixin):
    """
    Set of parameters for `permessage-bzip2` responded by server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension response for `permessage-bzip2` provided by a server to a client.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: A new instance of :class:`autobahn.compress.PerMessageBzip2Response`.
        :rtype: obj
        """
        client_max_compress_level = 0
        server_max_compress_level = 0

        for p in params:
            if len(params[p]) > 1:
                raise Exception(
                    f"multiple occurrence of extension parameter '{p}' for extension '{cls.EXTENSION_NAME}'"
                )

            val = params[p][0]

            if p == "client_max_compress_level":
                try:
                    val = int(val)
                except:
                    raise Exception(
                        f"illegal extension parameter value '{val}' for parameter '{p}' of extension '{cls.EXTENSION_NAME}'"
                    )
                if val not in PerMessageBzip2Mixin.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                    raise Exception(
                        f"illegal extension parameter value '{val}' for parameter '{p}' of extension '{cls.EXTENSION_NAME}'"
                    )
                else:
                    client_max_compress_level = val

            elif p == "server_max_compress_level":
                try:
                    val = int(val)
                except:
                    raise Exception(
                        f"illegal extension parameter value '{val}' for parameter '{p}' of extension '{cls.EXTENSION_NAME}'"
                    )
                if val not in PerMessageBzip2Mixin.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                    raise Exception(
                        f"illegal extension parameter value '{val}' for parameter '{p}' of extension '{cls.EXTENSION_NAME}'"
                    )
                else:
                    server_max_compress_level = val

            else:
                raise Exception(
                    f"illegal extension parameter '{p}' for extension '{cls.EXTENSION_NAME}'"
                )

        response = cls(client_max_compress_level, server_max_compress_level)
        return response

    def __init__(self, client_max_compress_level, server_max_compress_level):
        self.client_max_compress_level = client_max_compress_level
        self.server_max_compress_level = server_max_compress_level

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "client_max_compress_level": self.client_max_compress_level,
            "server_max_compress_level": self.server_max_compress_level,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            f"PerMessageBzip2Response(client_max_compress_level = {self.client_max_compress_level}, server_max_compress_level = {self.server_max_compress_level})"
        )


class PerMessageBzip2ResponseAccept(
    PerMessageCompressResponseAccept, PerMessageBzip2Mixin
):
    """
    Set of parameters with which to accept an `permessage-bzip2` response
    from a server by a client.
    """

    def __init__(self, response, compress_level=None):
        """

        :param response: The response being accepted.
        :type response: Instance of :class:`autobahn.compress.PerMessageBzip2Response`.
        :param compress_level: Override client ("client-to-server direction") compress level (this must be compatible with response).
        :type compress_level: int
        """
        if not isinstance(response, PerMessageBzip2Response):
            raise Exception(f"invalid type {type(response)} for response")

        self.response = response

        if compress_level is not None:
            if compress_level not in self.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                raise Exception(
                    f"invalid value {compress_level} for compress_level - permissible values {self.COMPRESS_LEVEL_PERMISSIBLE_VALUES}"
                )

            if (
                response.client_max_compress_level != 0
                and compress_level > response.client_max_compress_level
            ):
                raise Exception(
                    f"invalid value {compress_level} for compress_level - server requested lower maximum value"
                )

        self.compress_level = compress_level

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "response": self.response.__json__(),
            "compress_level": self.compress_level,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return f"PerMessageBzip2ResponseAccept(response = {self.response.__repr__()}, compress_level = {self.compress_level})"


class PerMessageBzip2(PerMessageCompress, PerMessageBzip2Mixin):
    """
    `permessage-bzip2` WebSocket extension processor.
    """

    DEFAULT_COMPRESS_LEVEL = 9

    @classmethod
    def create_from_response_accept(cls, is_server, accept):
        pmce = cls(
            is_server,
            accept.response.server_max_compress_level,
            (
                accept.compress_level
                if accept.compress_level is not None
                else accept.response.client_max_compress_level
            ),
        )
        return pmce

    @classmethod
    def create_from_offer_accept(cls, is_server, accept):
        pmce = cls(
            is_server,
            (
                accept.compress_level
                if accept.compress_level is not None
                else accept.offer.request_max_compress_level
            ),
            accept.request_max_compress_level,
        )
        return pmce

    def __init__(self, is_server, server_max_compress_level, client_max_compress_level):
        self._isServer = is_server
        self._compressor = None
        self._decompressor = None

        self.server_max_compress_level = (
            server_max_compress_level
            if server_max_compress_level != 0
            else self.DEFAULT_COMPRESS_LEVEL
        )
        self.client_max_compress_level = (
            client_max_compress_level
            if client_max_compress_level != 0
            else self.DEFAULT_COMPRESS_LEVEL
        )

    def __json__(self):
        return {
            "extension": self.EXTENSION_NAME,
            "isServer": self._isServer,
            "server_max_compress_level": self.server_max_compress_level,
            "client_max_compress_level": self.client_max_compress_level,
        }

    def __repr__(self):
        return (
            f"PerMessageBzip2(isServer = {self._isServer}, server_max_compress_level = {self.server_max_compress_level}, client_max_compress_level = {self.client_max_compress_level})"
        )

    def start_compress_message(self):
        if self._isServer:
            if self._compressor is None:
                self._compressor = bz2.BZ2Compressor(self.server_max_compress_level)
        else:
            if self._compressor is None:
                self._compressor = bz2.BZ2Compressor(self.client_max_compress_level)

    def compress_message_data(self, data):
        return self._compressor.compress(data)

    def end_compress_message(self):
        data = self._compressor.flush()

        # there seems to be no "flush without close stream", and after
        # full flush, compressor must not be reused
        self._compressor = None

        return data

    def start_decompress_message(self):
        if self._decompressor is None:
            self._decompressor = bz2.BZ2Decompressor()

    def decompress_message_data(self, data, max_output_len=None):
        if max_output_len is None:
            return self._decompressor.decompress(data)
        # BZ2Decompressor.decompress(data, max_length) returns at most
        # max_length bytes and buffers any excess internally. Cap at one octet
        # over the budget: if that yields more than max_output_len octets the
        # message is over budget and is rejected (matching deflate's
        # strictly-greater boundary). (needs_input is unreliable here - it also
        # goes False at end-of-stream, i.e. for an under-budget message.)
        data = self._decompressor.decompress(data, max(max_output_len, 0) + 1)
        if len(data) > max_output_len:
            raise PayloadExceededError(
                "WebSocket message exceeds decompression limit of "
                f"{max_output_len} octets"
            )
        return data

    def end_decompress_message(self):
        self._decompressor = None



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os

from cffi import FFI

# Try normal import first (works when package is installed or in editable mode)
try:
    from autobahn.nvx._compile_args import get_compile_args
except ImportError:
    # Fallback for CFFI build time (before package installation)
    # CFFI's setuptools integration runs builder modules via execfile() outside
    # of package context, so normal imports fail. Use importlib to dynamically
    # load the module from file path as a workaround.
    import importlib.util
    import sys

    _path = os.path.join(os.path.dirname(__file__), "_compile_args.py")
    spec = importlib.util.spec_from_file_location("autobahn.nvx._compile_args", _path)
    mod = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = mod
    spec.loader.exec_module(mod)
    get_compile_args = mod.get_compile_args

ffi = FFI()

ffi.cdef(
    """
    void* nvx_xormask_new (const uint8_t* mask);

    void nvx_xormask_reset (void* xormask);

    size_t nvx_xormask_pointer (void* xormask);

    void nvx_xormask_process (void* xormask, uint8_t* data, size_t length);

    void nvx_xormask_free (void* xormask);

    int nvx_xormask_set_impl(void* xormask, int impl);

    int nvx_xormask_get_impl(void* xormask);
"""
)

if "AUTOBAHN_USE_NVX" in os.environ and os.environ["AUTOBAHN_USE_NVX"] in ["1", "true"]:
    optional = False  # :noindex:
else:
    optional = True  # :noindex:

# Get appropriate compiler flags for this build context
# See autobahn.nvx._compile_args for details on architecture baseline selection
# and build context detection (wheel distribution vs. local source install)
extra_compile_args = get_compile_args()

with open(os.path.join(os.path.dirname(__file__), "_xormasker.c")) as fd:
    c_source = fd.read()
    ffi.set_source(
        "_nvx_xormasker",
        c_source,
        libraries=[],
        extra_compile_args=extra_compile_args,
        optional=optional,
    )


class XorMaskerNvx:
    """
    XOR masker using native NVX acceleration.
    :noindex:
    """

    __slots__ = ("ffi", "lib", "_masker", "_mask_buffer")

    def __init__(self, mask, use_simd=True):
        assert len(mask) == 4, "Mask must be exactly 4 bytes"

        from _nvx_xormasker import ffi as _ffi, lib as _lib

        self.ffi = _ffi
        self.lib = _lib

        # Keep mask buffer alive for the lifetime of the masker
        self._mask_buffer = self.ffi.new("uint8_t[4]", mask)
        self._masker = self.ffi.gc(
            self.lib.nvx_xormask_new(self._mask_buffer), self.lib.nvx_xormask_free
        )

        # Set implementation: 1=Simple (scalar), 2=SSE2 (SIMD)
        if use_simd:
            self.lib.nvx_xormask_set_impl(self._masker, 2)  # XOR_MASKER_SSE2
        else:
            self.lib.nvx_xormask_set_impl(self._masker, 1)  # XOR_MASKER_SIMPLE

    def pointer(self):
        return self.lib.nvx_xormask_pointer(self._masker)

    def reset(self):
        self.lib.nvx_xormask_reset(self._masker)

    def process(self, data):
        # Create a mutable copy of the data
        data_len = len(data)
        data_buffer = self.ffi.new("uint8_t[]", data_len)
        self.ffi.memmove(data_buffer, data, data_len)

        # Process in-place
        self.lib.nvx_xormask_process(self._masker, data_buffer, data_len)

        # Convert back to bytes
        return bytes(self.ffi.buffer(data_buffer, data_len))


class XorMaskerSimple(XorMaskerNvx):
    """
    Simple scalar XOR masker for small payloads.
    :noindex:
    """

    def __init__(self, mask):
        super().__init__(mask, use_simd=False)


class XorMaskerShifted1(XorMaskerNvx):
    """
    SIMD-optimized XOR masker for large payloads.
    :noindex:
    """

    def __init__(self, mask):
        super().__init__(mask, use_simd=True)


def create_xor_masker(mask, length=None):
    """
    Factory function to create XOR masker with optimal implementation.

    :param mask: 4-byte masking key
    :param length: Payload length hint - uses scalar for < 128 bytes, SIMD for >= 128
    :return: XorMaskerSimple or XorMaskerShifted1 instance
    """
    if length is None or length < 128:
        return XorMaskerSimple(mask)
    else:
        return XorMaskerShifted1(mask)


if __name__ == "__main__":
    ffi.compile()



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest
from unittest.mock import patch

from zope.interface import implementer

from twisted.internet.interfaces import IReactorTime


@implementer(IReactorTime)
class FakeReactor:
    """
    This just fakes out enough reactor methods so .run() can work.
    """

    stop_called = False

    def __init__(self, to_raise):
        self.stop_called = False
        self.to_raise = to_raise
        self.delayed = []

    def run(self, *args, **kw):
        raise self.to_raise

    def stop(self):
        self.stop_called = True

    def callLater(self, delay, func, *args, **kwargs):
        self.delayed.append((delay, func, args, kwargs))

    def connectTCP(self, *args, **kw):
        raise RuntimeError("ConnectTCP shouldn't get called")


class TestWampTwistedRunner(unittest.TestCase):
    # XXX should figure out *why* but the test_protocol timeout
    # tests fail if we *don't* patch out this txaio stuff. So,
    # presumably it's messing up some global state that both tests
    # implicitly depend on ...

    @patch("txaio.use_twisted")
    @patch("txaio.start_logging")
    @patch("txaio.config")
    def test_connect_error(self, *args):
        """
        Ensure the runner doesn't swallow errors and that it exits the
        reactor properly if there is one.
        """
        try:
            from autobahn.twisted.wamp import ApplicationRunner

            # the 'reactor' member doesn't exist until we import it
            from twisted.internet import reactor  # noqa: F401
            from twisted.internet.error import ConnectionRefusedError
        except ImportError:
            raise unittest.SkipTest("No twisted")

        runner = ApplicationRunner("ws://localhost:1", "realm")
        exception = ConnectionRefusedError("It's a trap!")

        with patch("twisted.internet.reactor", FakeReactor(exception)) as mockreactor:
            self.assertRaises(
                ConnectionRefusedError,
                # pass a no-op session-creation method
                runner.run,
                lambda _: None,
                start_reactor=True,
            )
            self.assertTrue(mockreactor.stop_called)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import decimal
import os
import random
import unittest
from decimal import Decimal

from autobahn.wamp import message, role, serializer


def generate_test_messages():
    """
    List of WAMP test message used for serializers. Expand this if you add more
    options or messages.

    This list of WAMP message does not contain any binary app payloads!
    """
    some_bytes = os.urandom(32)
    some_unicode = "\u3053\u3093\u306b\u3061\u306f\u4e16\u754c"

    some_uri = "com.myapp.foobar"
    some_unicode_uri = "com.myapp.\u4f60\u597d\u4e16\u754c.baz"

    some_args = [
        1,
        2,
        3,
        "hello",
        some_bytes,
        some_unicode,
        {"foo": 23, "bar": "hello", "baz": some_bytes, "moo": some_unicode},
    ]
    some_kwargs = {
        "foo": 23,
        "bar": "hello",
        "baz": some_bytes,
        "moo": some_unicode,
        "arr": some_args,
    }

    msgs = [
        message.Hello("realm1", {"subscriber": role.RoleSubscriberFeatures()}),
        message.Hello("realm1", {"publisher": role.RolePublisherFeatures()}),
        message.Hello("realm1", {"caller": role.RoleCallerFeatures()}),
        message.Hello("realm1", {"callee": role.RoleCalleeFeatures()}),
        message.Hello(
            "realm1",
            {
                "subscriber": role.RoleSubscriberFeatures(),
                "publisher": role.RolePublisherFeatures(),
                "caller": role.RoleCallerFeatures(),
                "callee": role.RoleCalleeFeatures(),
            },
        ),
        message.Goodbye(),
        message.Yield(123456),
        message.Yield(123456, args=some_args),
        message.Yield(123456, args=[], kwargs=some_kwargs),
        message.Yield(123456, args=some_args, kwargs=some_kwargs),
        message.Yield(123456, progress=True),
        message.Interrupt(123456),
        message.Interrupt(123456, mode=message.Interrupt.KILL),
        message.Invocation(123456, 789123),
        message.Invocation(123456, 789123, args=some_args),
        message.Invocation(123456, 789123, args=[], kwargs=some_kwargs),
        message.Invocation(123456, 789123, args=some_args, kwargs=some_kwargs),
        message.Invocation(123456, 789123, timeout=10000),
        message.Result(123456),
        message.Result(123456, args=some_args),
        message.Result(123456, args=[], kwargs=some_kwargs),
        message.Result(123456, args=some_args, kwargs=some_kwargs),
        message.Result(123456, progress=True),
        message.Cancel(123456),
        message.Cancel(123456, mode=message.Cancel.KILL),
        message.Call(123456, some_uri),
        message.Call(123456, some_uri, args=some_args),
        message.Call(123456, some_uri, args=[], kwargs=some_kwargs),
        message.Call(123456, some_uri, args=some_args, kwargs=some_kwargs),
        message.Call(123456, some_uri, timeout=10000),
        message.Call(123456, some_unicode_uri),
        message.Call(123456, some_unicode_uri, args=some_args),
        message.Call(123456, some_unicode_uri, args=[], kwargs=some_kwargs),
        message.Call(123456, some_unicode_uri, args=some_args, kwargs=some_kwargs),
        message.Call(123456, some_unicode_uri, timeout=10000),
        message.Unregistered(123456),
        message.Unregister(123456, 789123),
        message.Registered(123456, 789123),
        message.Register(123456, some_uri),
        message.Register(123456, some_uri, match="prefix"),
        message.Register(123456, some_uri, invoke="roundrobin"),
        message.Register(123456, some_unicode_uri),
        message.Register(123456, some_unicode_uri, match="prefix"),
        message.Register(123456, some_unicode_uri, invoke="roundrobin"),
        message.Event(123456, 789123),
        message.Event(123456, 789123, args=some_args),
        message.Event(123456, 789123, args=[], kwargs=some_kwargs),
        message.Event(123456, 789123, args=some_args, kwargs=some_kwargs),
        message.Event(123456, 789123, publisher=300),
        message.Published(123456, 789123),
        message.Publish(123456, some_uri),
        message.Publish(123456, some_uri, args=some_args),
        message.Publish(123456, some_uri, args=[], kwargs=some_kwargs),
        message.Publish(123456, some_uri, args=some_args, kwargs=some_kwargs),
        message.Publish(
            123456, some_uri, exclude_me=False, exclude=[300], eligible=[100, 200, 300]
        ),
        message.Publish(123456, some_unicode_uri),
        message.Publish(123456, some_unicode_uri, args=some_args),
        message.Publish(123456, some_unicode_uri, args=[], kwargs=some_kwargs),
        message.Publish(123456, some_unicode_uri, args=some_args, kwargs=some_kwargs),
        message.Publish(
            123456,
            some_unicode_uri,
            exclude_me=False,
            exclude=[300],
            eligible=[100, 200, 300],
        ),
        message.Unsubscribed(123456),
        message.Unsubscribe(123456, 789123),
        message.Subscribed(123456, 789123),
        message.Subscribe(123456, some_uri),
        message.Subscribe(123456, some_uri, match=message.Subscribe.MATCH_PREFIX),
        message.Subscribe(123456, some_unicode_uri),
        message.Subscribe(
            123456, some_unicode_uri, match=message.Subscribe.MATCH_PREFIX
        ),
        message.Error(message.Call.MESSAGE_TYPE, 123456, some_uri),
        message.Error(message.Call.MESSAGE_TYPE, 123456, some_uri, args=some_args),
        message.Error(
            message.Call.MESSAGE_TYPE, 123456, some_uri, args=[], kwargs=some_kwargs
        ),
        message.Error(
            message.Call.MESSAGE_TYPE,
            123456,
            some_uri,
            args=some_args,
            kwargs=some_kwargs,
        ),
        message.Error(message.Call.MESSAGE_TYPE, 123456, some_unicode_uri),
        message.Error(
            message.Call.MESSAGE_TYPE, 123456, some_unicode_uri, args=some_args
        ),
        message.Error(
            message.Call.MESSAGE_TYPE,
            123456,
            some_unicode_uri,
            args=[],
            kwargs=some_kwargs,
        ),
        message.Error(
            message.Call.MESSAGE_TYPE,
            123456,
            some_unicode_uri,
            args=some_args,
            kwargs=some_kwargs,
        ),
        message.Result(123456),
        message.Result(123456, args=some_args),
        message.Result(123456, args=some_args, kwargs=some_kwargs),
    ]
    return [(False, msg) for msg in msgs]


def generate_test_messages_binary():
    """
    Generate WAMP test messages which contain binary app payloads.

    With the JSON serializer, this currently only works on Python 3 (both CPython3 and PyPy3),
    because even on Python 3, we need to patch the stdlib JSON, and on Python 2, the patching
    would be even hackier.
    """
    msgs = []
    for binary in [
        b"",
        b"\x00",
        b"\30",
        os.urandom(4),
        os.urandom(16),
        os.urandom(128),
        os.urandom(256),
        os.urandom(512),
        os.urandom(1024),
    ]:
        msgs.append(message.Event(123456, 789123, args=[binary]))
        msgs.append(
            message.Event(123456, 789123, args=[binary], kwargs={"foo": binary})
        )
    return [(True, msg) for msg in msgs]


def create_serializers(decimal_support=False):
    _serializers = []

    _serializers.append(serializer.JsonSerializer(use_decimal_from_str=decimal_support))
    _serializers.append(
        serializer.JsonSerializer(batched=True, use_decimal_from_str=decimal_support)
    )

    _serializers.append(serializer.CBORSerializer())
    _serializers.append(serializer.CBORSerializer(batched=True))

    if not decimal_support:
        # builtins.OverflowError: Integer value out of range
        _serializers.append(serializer.MsgPackSerializer())
        _serializers.append(serializer.MsgPackSerializer(batched=True))

        # UBJSON (bjdata-backed) is optional: only present when the
        # `autobahn[serialization]` extra (bjdata) is installed.
        if hasattr(serializer, "UBJSONSerializer"):
            _serializers.append(serializer.UBJSONSerializer())
            _serializers.append(serializer.UBJSONSerializer(batched=True))

    # FIXME: implement full FlatBuffers serializer for WAMP
    # WAMP-FlatBuffers currently only supports Python 3
    # _serializers.append(serializer.FlatBuffersSerializer())
    # _serializers.append(serializer.FlatBuffersSerializer(batched=True))

    return _serializers


class TestFlatBuffersSerializer(unittest.TestCase):
    def test_basic(self):
        messages = [
            message.Event(
                123456,
                789123,
                args=[1, 2, 3],
                kwargs={"foo": 23, "bar": "hello"},
                publisher=666,
                retained=True,
            ),
            message.Publish(
                123456,
                "com.example.topic1",
                args=[1, 2, 3],
                kwargs={"foo": 23, "bar": "hello"},
                retain=True,
            ),
        ]

        ser = serializer.FlatBuffersSerializer()

        # from pprint import pprint

        for msg in messages:
            # serialize message
            payload, binary = ser.serialize(msg)

            # unserialize message again
            msg2 = ser.unserialize(payload, binary)[0]

            # pprint(msg.marshal())
            # pprint(msg2.marshal())

            # must be equal: message roundtrips via the serializer
            self.assertEqual(msg, msg2)
            # self.assertEqual(msg.subscription, msg2.subscription)
            # self.assertEqual(msg.publication, msg2.publication)


class TestDecimalSerializer(unittest.TestCase):
    """
    binary fixed-point
    binary floating-point:          float (float32), double (float64)
    decimal floating-point:         decimal128, decimal256
    decimal fixed-point:            NUMERIC(precision, scale)
    decimal arbitrary precision:    NUMERIC, decimal.Decimal

    https://developer.nvidia.com/blog/implementing-high-precision-decimal-arithmetic-with-cuda-int128/
    https://github.com/johnmcfarlane/cnl
    """

    def setUp(self) -> None:
        self._test_serializers = create_serializers(decimal_support=True)

        # enough for decimal256 precision arithmetic (76 significand decimal digits)
        decimal.getcontext().prec = 76

        self._test_messages_no_dec = [
            (
                True,
                {
                    "a": random.random(),
                    "b": random.randint(0, 2**53),
                    "c": random.randint(0, 2**64),
                    "d": random.randint(0, 2**128),
                    "e": random.randint(0, 2**256),
                    # float64: 52 binary digits, precision of 15-17 significant decimal digits
                    "f": 0.12345678901234567,
                    "g": 0.8765432109876545,
                    "y": os.urandom(8),
                    "z": [
                        -1,
                        0,
                        1,
                        True,
                        None,
                        0.12345678901234567,
                        0.8765432109876545,
                        os.urandom(8),
                    ],
                },
            )
        ]
        self._test_messages_dec = [
            (
                True,
                {
                    "a": random.random(),
                    "b": random.randint(0, 2**53),
                    "c": random.randint(0, 2**64),
                    "d": random.randint(0, 2**128),
                    "e": random.randint(0, 2**256),
                    # float64: 52 binary digits, precision of 15-17 significant decimal digits
                    "f": 0.12345678901234567,
                    "g": 0.8765432109876545,
                    # decimal128: precision of 38 significant decimal digits
                    "h": Decimal("0.1234567890123456789012345678901234567"),
                    "i": Decimal("0.8765432109876543210987654321098765434"),
                    # decimal256: precision of 76 significant decimal digits
                    "j": Decimal(
                        "0.123456789012345678901234567890123456701234567890123456789012345678901234567"
                    ),
                    "k": Decimal(
                        "0.876543210987654321098765432109876543298765432109876543210987654321098765434"
                    ),
                    "y": os.urandom(8),
                    "z": [
                        -1,
                        0,
                        1,
                        True,
                        None,
                        0.12345678901234567,
                        0.8765432109876545,
                        Decimal("0.1234567890123456789012345678901234567"),
                        Decimal("0.8765432109876543210987654321098765434"),
                        Decimal(
                            "0.123456789012345678901234567890123456701234567890123456789012345678901234567"
                        ),
                        Decimal(
                            "0.876543210987654321098765432109876543298765432109876543210987654321098765434"
                        ),
                        os.urandom(8),
                    ],
                },
            )
        ]

    def test_json_no_decimal(self):
        """
        Test without ``use_decimal_from_str`` feature of JSON object serializer.
        """
        ser = serializer.JsonObjectSerializer(use_decimal_from_str=False)
        for contains_binary, obj in self._test_messages_no_dec:
            _obj = ser.unserialize(ser.serialize(obj))[0]
            self.assertEqual(obj, _obj)
            self.assertEqual(1.0000000000000002, _obj["f"] + _obj["g"])

    def test_json_decimal(self):
        """
        Test ``use_decimal_from_str`` feature of JSON object serializer.
        """
        ser = serializer.JsonObjectSerializer(use_decimal_from_str=True)
        for contains_binary, obj in self._test_messages_dec:
            _obj = ser.unserialize(ser.serialize(obj))[0]
            self.assertEqual(obj, _obj)
            self.assertEqual(1.0000000000000002, _obj["f"] + _obj["g"])
            self.assertEqual(
                Decimal("1.0000000000000000000000000000000000001"),
                _obj["h"] + _obj["i"],
            )
            self.assertEqual(
                Decimal(
                    "1.000000000000000000000000000000000000000000000000000000000000000000000000001"
                ),
                _obj["j"] + _obj["k"],
            )

    def test_roundtrip_msg(self):
        for wamp_ser in self._test_serializers:
            ser = wamp_ser._serializer
            for contains_binary, msg in (
                self._test_messages_no_dec + self._test_messages_dec
            ):
                payload = ser.serialize(msg)
                msg2 = ser.unserialize(payload)
                self.assertEqual(msg, msg2[0])

    def test_crosstrip_msg(self):
        for wamp_ser1 in self._test_serializers:
            ser1 = wamp_ser1._serializer
            for contains_binary, msg in (
                self._test_messages_no_dec + self._test_messages_dec
            ):
                payload1 = ser1.serialize(msg)
                msg1 = ser1.unserialize(payload1)
                msg1 = msg1[0]
                for wamp_ser2 in self._test_serializers:
                    ser2 = wamp_ser2._serializer
                    payload2 = ser2.serialize(msg1)
                    msg2 = ser2.unserialize(payload2)
                    msg2 = msg2[0]
                    self.assertEqual(msg, msg2)
                    # print(ser1, len(payload1), ser2, len(payload2))


class TestSerializer(unittest.TestCase):
    def setUp(self):
        self._test_messages = generate_test_messages() + generate_test_messages_binary()
        self._test_serializers = create_serializers()
        # print('Testing WAMP serializers {} with {} WAMP test messages'.format([ser.SERIALIZER_ID for ser in self._test_serializers], len(self._test_messages)))

    def test_deep_equal_msg(self):
        """
        Test deep object equality assert (because I am paranoid).
        """
        v = os.urandom(10)
        o1 = [
            1,
            2,
            {"foo": "bar", "bar": v, "baz": [9, 3, 2], "goo": {"moo": [1, 2, 3]}},
            v,
        ]
        o2 = [
            1,
            2,
            {"goo": {"moo": [1, 2, 3]}, "bar": v, "baz": [9, 3, 2], "foo": "bar"},
            v,
        ]
        self.assertEqual(o1, o2)

    def test_roundtrip_msg(self):
        """
        Test round-tripping over each serializer.
        """
        for ser in self._test_serializers:
            for contains_binary, msg in self._test_messages:
                # serialize message
                payload, binary = ser.serialize(msg)

                # unserialize message again
                msg2 = ser.unserialize(payload, binary)

                # must be equal: message roundtrips via the serializer
                self.assertEqual([msg], msg2)

    def test_crosstrip_msg(self):
        """
        Test cross-tripping over 2 serializers (as is done by WAMP routers).
        """
        for ser1 in self._test_serializers:
            for contains_binary, msg in self._test_messages:
                # serialize message
                payload, binary = ser1.serialize(msg)

                # unserialize message again
                msg1 = ser1.unserialize(payload, binary)
                msg1 = msg1[0]

                for ser2 in self._test_serializers:
                    # serialize message
                    payload, binary = ser2.serialize(msg1)

                    # unserialize message again
                    msg2 = ser2.unserialize(payload, binary)

                    # must be equal: message crosstrips via
                    # the serializers ser1 -> ser2
                    self.assertEqual([msg], msg2)

    def test_cache_msg(self):
        """
        Test message serialization caching.
        """
        for contains_binary, msg in self._test_messages:
            # message serialization cache is initially empty
            self.assertEqual(msg._serialized, {})

            for ser in self._test_serializers:
                # verify message serialization is not yet cached
                self.assertFalse(ser._serializer in msg._serialized)
                payload, binary = ser.serialize(msg)

                # now the message serialization must be cached
                self.assertTrue(ser._serializer in msg._serialized)
                self.assertEqual(msg._serialized[ser._serializer], payload)

                # and after resetting the serialization cache, message
                # serialization is gone
                msg.uncache()
                self.assertFalse(ser._serializer in msg._serialized)

    def test_initial_stats(self):
        """
        Test initial serializer stats are indeed empty.
        """
        for ser in self._test_serializers:
            stats = ser.stats(details=True)

            self.assertEqual(stats["serialized"]["bytes"], 0)
            self.assertEqual(stats["serialized"]["messages"], 0)
            self.assertEqual(stats["serialized"]["rated_messages"], 0)

            self.assertEqual(stats["unserialized"]["bytes"], 0)
            self.assertEqual(stats["unserialized"]["messages"], 0)
            self.assertEqual(stats["unserialized"]["rated_messages"], 0)

    def test_serialize_stats(self):
        """
        Test serializer stats are non-empty after serializing/unserializing messages.
        """
        for ser in self._test_serializers:
            for contains_binary, msg in self._test_messages:
                # serialize message
                payload, binary = ser.serialize(msg)

                # unserialize message again
                ser.unserialize(payload, binary)

            stats = ser.stats(details=False)

            self.assertTrue(stats["bytes"] > 0)
            self.assertTrue(stats["messages"] > 0)
            self.assertTrue(stats["rated_messages"] > 0)

    def test_serialize_stats_with_details(self):
        """
        Test serializer stats - with details - are non-empty after serializing/unserializing messages.
        """
        for ser in self._test_serializers:
            for contains_binary, msg in self._test_messages:
                # serialize message
                payload, binary = ser.serialize(msg)

                # unserialize message again
                ser.unserialize(payload, binary)

            stats = ser.stats(details=True)

            # {'serialized': {'bytes': 7923, 'messages': 59, 'rated_messages': 69}, 'unserialized': {'bytes': 7923, 'messages': 59, 'rated_messages': 69}}
            # print(stats)

            self.assertTrue(stats["serialized"]["bytes"] > 0)
            self.assertTrue(stats["serialized"]["messages"] > 0)
            self.assertTrue(stats["serialized"]["rated_messages"] > 0)

            self.assertTrue(stats["unserialized"]["bytes"] > 0)
            self.assertTrue(stats["unserialized"]["messages"] > 0)
            self.assertTrue(stats["unserialized"]["rated_messages"] > 0)

            self.assertEqual(
                stats["serialized"]["bytes"], stats["unserialized"]["bytes"]
            )
            self.assertEqual(
                stats["serialized"]["messages"], stats["unserialized"]["messages"]
            )
            self.assertEqual(
                stats["serialized"]["rated_messages"],
                stats["unserialized"]["rated_messages"],
            )

    def test_reset_stats(self):
        """
        Test serializer stats are reset after fetching stats - depending on option.
        """
        for ser in self._test_serializers:
            for contains_binary, msg in self._test_messages:
                # serialize message
                payload, binary = ser.serialize(msg)

                # unserialize message again
                ser.unserialize(payload, binary)

            ser.stats()
            stats = ser.stats(details=True)

            self.assertEqual(stats["serialized"]["bytes"], 0)
            self.assertEqual(stats["serialized"]["messages"], 0)
            self.assertEqual(stats["serialized"]["rated_messages"], 0)

            self.assertEqual(stats["unserialized"]["bytes"], 0)
            self.assertEqual(stats["unserialized"]["messages"], 0)
            self.assertEqual(stats["unserialized"]["rated_messages"], 0)

    def test_auto_stats(self):
        """
        Test serializer stats are non-empty after serializing/unserializing messages.
        """
        for ser in self._test_serializers:

            def on_stats(stats):
                self.assertTrue(stats["bytes"] > 0)
                self.assertTrue(stats["messages"] > 0)
                self.assertTrue(stats["rated_messages"] > 0)

            ser.set_stats_autoreset(10, 0, on_stats)

            for contains_binary, msg in self._test_messages:
                # serialize message
                payload, binary = ser.serialize(msg)

                # unserialize message again
                ser.unserialize(payload, binary)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import json

from autobahn import util
from autobahn.wamp.exception import ProtocolError

__all__ = (
    "DEFAULT_CLIENT_ROLES",
    "ROLE_NAME_TO_CLASS",
    "RoleBrokerFeatures",
    "RoleCalleeFeatures",
    "RoleCallerFeatures",
    "RoleDealerFeatures",
    "RoleFeatures",
    "RolePublisherFeatures",
    "RoleSubscriberFeatures",
)


class RoleFeatures(util.EqualityMixin):
    """
    Base class for WAMP role features.
    """

    ROLE = None

    def __str__(self):
        return json.dumps(self.__dict__)

    def __repr__(self):
        configured_options = {}
        for k, v in self.__dict__.items():
            if v is not None:
                configured_options[k] = v
        return "{}({})".format(
            self.ROLE,
            ", ".join([k + "=" + str(v) for k, v in configured_options.items()]),
        )

    def _check_all_bool(self):
        # check feature attributes
        for k in self.__dict__:
            if not k.startswith("_") and k != "ROLE":
                if getattr(self, k) is not None and type(getattr(self, k)) != bool:
                    raise ProtocolError(
                        f"invalid type {getattr(self, k)} for feature '{k}' for role '{self.ROLE}'"
                    )


class RoleBrokerFeatures(RoleFeatures):
    """
    WAMP broker role features.
    """

    ROLE = "broker"

    def __init__(
        self,
        publisher_identification=None,
        publication_trustlevels=None,
        pattern_based_subscription=None,
        session_meta_api=None,
        subscription_meta_api=None,
        subscriber_blackwhite_listing=None,
        publisher_exclusion=None,
        subscription_revocation=None,
        event_history=None,
        payload_transparency=None,
        x_acknowledged_event_delivery=None,
        payload_encryption_cryptobox=None,
        event_retention=None,
        **kwargs,
    ):
        self.publisher_identification = publisher_identification
        self.publication_trustlevels = publication_trustlevels
        self.pattern_based_subscription = pattern_based_subscription
        self.session_meta_api = session_meta_api
        self.subscription_meta_api = subscription_meta_api
        self.subscriber_blackwhite_listing = subscriber_blackwhite_listing
        self.publisher_exclusion = publisher_exclusion
        self.subscription_revocation = subscription_revocation
        self.event_history = event_history
        self.payload_transparency = payload_transparency
        self.x_acknowledged_event_delivery = x_acknowledged_event_delivery
        self.payload_encryption_cryptobox = payload_encryption_cryptobox
        self.event_retention = event_retention
        self._check_all_bool()


class RoleSubscriberFeatures(RoleFeatures):
    """
    WAMP subscriber role features.
    """

    ROLE = "subscriber"

    def __init__(
        self,
        publisher_identification=None,
        publication_trustlevels=None,
        pattern_based_subscription=None,
        subscription_revocation=None,
        event_history=None,
        payload_transparency=None,
        payload_encryption_cryptobox=None,
        **kwargs,
    ):
        self.publisher_identification = publisher_identification
        self.publication_trustlevels = publication_trustlevels
        self.pattern_based_subscription = pattern_based_subscription
        self.subscription_revocation = subscription_revocation
        self.event_history = event_history
        self.payload_transparency = payload_transparency
        self.payload_encryption_cryptobox = payload_encryption_cryptobox
        self._check_all_bool()


class RolePublisherFeatures(RoleFeatures):
    """
    WAMP publisher role features.
    """

    ROLE = "publisher"

    def __init__(
        self,
        publisher_identification=None,
        subscriber_blackwhite_listing=None,
        publisher_exclusion=None,
        payload_transparency=None,
        x_acknowledged_event_delivery=None,
        payload_encryption_cryptobox=None,
        **kwargs,
    ):
        self.publisher_identification = publisher_identification
        self.subscriber_blackwhite_listing = subscriber_blackwhite_listing
        self.publisher_exclusion = publisher_exclusion
        self.payload_transparency = payload_transparency
        self.x_acknowledged_event_delivery = x_acknowledged_event_delivery
        self.payload_encryption_cryptobox = payload_encryption_cryptobox
        self._check_all_bool()


class RoleDealerFeatures(RoleFeatures):
    """
    WAMP dealer role features.
    """

    ROLE = "dealer"

    def __init__(
        self,
        caller_identification=None,
        call_trustlevels=None,
        pattern_based_registration=None,
        session_meta_api=None,
        registration_meta_api=None,
        shared_registration=None,
        call_timeout=None,
        call_canceling=None,
        progressive_call_results=None,
        registration_revocation=None,
        payload_transparency=None,
        testament_meta_api=None,
        payload_encryption_cryptobox=None,
        **kwargs,
    ):
        self.caller_identification = caller_identification
        self.call_trustlevels = call_trustlevels
        self.pattern_based_registration = pattern_based_registration
        self.session_meta_api = session_meta_api
        self.registration_meta_api = registration_meta_api
        self.shared_registration = shared_registration
        self.call_timeout = call_timeout
        self.call_canceling = call_canceling
        self.progressive_call_results = progressive_call_results
        self.registration_revocation = registration_revocation
        self.payload_transparency = payload_transparency
        self.testament_meta_api = testament_meta_api
        self.payload_encryption_cryptobox = payload_encryption_cryptobox
        self._check_all_bool()


class RoleCallerFeatures(RoleFeatures):
    """
    WAMP caller role features.
    """

    ROLE = "caller"

    def __init__(
        self,
        caller_identification=None,
        call_timeout=None,
        call_canceling=None,
        progressive_call_results=None,
        payload_transparency=None,
        payload_encryption_cryptobox=None,
        **kwargs,
    ):
        self.caller_identification = caller_identification
        self.call_timeout = call_timeout
        self.call_canceling = call_canceling
        self.progressive_call_results = progressive_call_results
        self.payload_transparency = payload_transparency
        self.payload_encryption_cryptobox = payload_encryption_cryptobox
        self._check_all_bool()


class RoleCalleeFeatures(RoleFeatures):
    """
    WAMP callee role features.
    """

    ROLE = "callee"

    def __init__(
        self,
        caller_identification=None,
        call_trustlevels=None,
        pattern_based_registration=None,
        shared_registration=None,
        call_timeout=None,
        call_canceling=None,
        progressive_call_results=None,
        registration_revocation=None,
        payload_transparency=None,
        payload_encryption_cryptobox=None,
        **kwargs,
    ):
        self.caller_identification = caller_identification
        self.call_trustlevels = call_trustlevels
        self.pattern_based_registration = pattern_based_registration
        self.shared_registration = shared_registration
        self.call_timeout = call_timeout
        self.call_canceling = call_canceling
        self.progressive_call_results = progressive_call_results
        self.registration_revocation = registration_revocation
        self.payload_transparency = payload_transparency
        self.payload_encryption_cryptobox = payload_encryption_cryptobox
        self._check_all_bool()


# map of role names to role class
ROLE_NAME_TO_CLASS = {
    "broker": RoleBrokerFeatures,
    "subscriber": RoleSubscriberFeatures,
    "publisher": RolePublisherFeatures,
    "dealer": RoleDealerFeatures,
    "caller": RoleCallerFeatures,
    "callee": RoleCalleeFeatures,
}


# default role features for client roles supported
DEFAULT_CLIENT_ROLES = {
    "subscriber": RoleSubscriberFeatures(
        publisher_identification=True,
        pattern_based_subscription=True,
        subscription_revocation=True,
        payload_transparency=True,
        payload_encryption_cryptobox=True,
    ),
    "publisher": RolePublisherFeatures(
        publisher_identification=True,
        subscriber_blackwhite_listing=True,
        publisher_exclusion=True,
        payload_transparency=True,
        x_acknowledged_event_delivery=True,
        payload_encryption_cryptobox=True,
    ),
    "caller": RoleCallerFeatures(
        caller_identification=True,
        progressive_call_results=True,
        payload_transparency=True,
        payload_encryption_cryptobox=True,
        call_canceling=True,
    ),
    "callee": RoleCalleeFeatures(
        caller_identification=True,
        pattern_based_registration=True,
        shared_registration=True,
        progressive_call_results=True,
        registration_revocation=True,
        payload_transparency=True,
        payload_encryption_cryptobox=True,
        call_canceling=True,
    ),
}



########################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
########################################

import sys
import traceback

import txaio

txaio.use_twisted()

from twisted.internet.error import ReactorAlreadyInstalledError
from twisted.python import reflect

__all__ = ("current_reactor_klass", "install_optimal_reactor", "install_reactor")


def current_reactor_klass():
    """
    Return class name of currently installed Twisted reactor or None.
    """
    if "twisted.internet.reactor" in sys.modules:
        current_reactor = reflect.qual(
            sys.modules["twisted.internet.reactor"].__class__
        ).split(".")[-1]
    else:
        current_reactor = None
    return current_reactor


def install_optimal_reactor(require_optimal_reactor=True):
    """
    Try to install the optimal Twisted reactor for this platform:

    - Linux:   epoll
    - BSD/OSX: kqueue
    - Windows: iocp
    - Other:   select

    Notes:

    - This function exists, because the reactor types selected based on platform
      in `twisted.internet.default` are different from here.
    - The imports are inlined, because the Twisted code base is notorious for
      importing the reactor as a side-effect of merely importing. Hence we postpone
      all importing.

    See: http://twistedmatrix.com/documents/current/core/howto/choosing-reactor.html#reactor-functionality

    :param require_optimal_reactor: If ``True`` and the desired reactor could not be installed,
        raise ``ReactorAlreadyInstalledError``, else fallback to another reactor.
    :type require_optimal_reactor: bool

    :returns: The Twisted reactor in place (`twisted.internet.reactor`).
    """
    log = txaio.make_logger()

    # determine currently installed reactor, if any
    #
    current_reactor = current_reactor_klass()

    # depending on platform, install optimal reactor
    #
    if "bsd" in sys.platform or sys.platform.startswith("darwin"):
        # *BSD and MacOSX
        #
        if current_reactor != "KQueueReactor":
            if current_reactor is None:
                try:
                    from twisted.internet import kqreactor

                    kqreactor.install()
                except:
                    log.warn(
                        "Running on *BSD or MacOSX, but cannot install kqueue Twisted reactor: {tb}",
                        tb=traceback.format_exc(),
                    )
                else:
                    log.debug(
                        "Running on *BSD or MacOSX and optimal reactor (kqueue) was installed."
                    )
            else:
                log.warn(
                    "Running on *BSD or MacOSX, but cannot install kqueue Twisted reactor, because another reactor ({klass}) is already installed.",
                    klass=current_reactor,
                )
                if require_optimal_reactor:
                    raise ReactorAlreadyInstalledError()
        else:
            log.debug(
                "Running on *BSD or MacOSX and optimal reactor (kqueue) already installed."
            )

    elif sys.platform in ["win32"]:
        # Windows
        #
        if current_reactor != "IOCPReactor":
            if current_reactor is None:
                try:
                    from twisted.internet.iocpreactor import reactor as iocpreactor

                    iocpreactor.install()
                except:
                    log.warn(
                        "Running on Windows, but cannot install IOCP Twisted reactor: {tb}",
                        tb=traceback.format_exc(),
                    )
                else:
                    log.debug(
                        "Running on Windows and optimal reactor (ICOP) was installed."
                    )
            else:
                log.warn(
                    "Running on Windows, but cannot install IOCP Twisted reactor, because another reactor ({klass}) is already installed.",
                    klass=current_reactor,
                )
                if require_optimal_reactor:
                    raise ReactorAlreadyInstalledError()
        else:
            log.debug(
                "Running on Windows and optimal reactor (ICOP) already installed."
            )

    elif sys.platform.startswith("linux"):
        # Linux
        #
        if current_reactor != "EPollReactor":
            if current_reactor is None:
                try:
                    from twisted.internet import epollreactor

                    epollreactor.install()
                except:
                    log.warn(
                        "Running on Linux, but cannot install Epoll Twisted reactor: {tb}",
                        tb=traceback.format_exc(),
                    )
                else:
                    log.debug(
                        "Running on Linux and optimal reactor (epoll) was installed."
                    )
            else:
                log.warn(
                    "Running on Linux, but cannot install Epoll Twisted reactor, because another reactor ({klass}) is already installed.",
                    klass=current_reactor,
                )
                if require_optimal_reactor:
                    raise ReactorAlreadyInstalledError()
        else:
            log.debug("Running on Linux and optimal reactor (epoll) already installed.")

    else:
        # Other platform
        #
        if current_reactor != "SelectReactor":
            if current_reactor is None:
                try:
                    from twisted.internet import selectreactor

                    selectreactor.install()
                    # from twisted.internet import default as defaultreactor
                    # defaultreactor.install()
                except:
                    log.warn(
                        'Running on "{platform}", but cannot install Select Twisted reactor: {tb}',
                        tb=traceback.format_exc(),
                        platform=sys.platform,
                    )
                else:
                    log.debug(
                        'Running on "{platform}" and optimal reactor (Select) was installed.',
                        platform=sys.platform,
                    )
            else:
                log.warn(
                    'Running on "{platform}", but cannot install Select Twisted reactor, because another reactor ({klass}) is already installed.',
                    klass=current_reactor,
                    platform=sys.platform,
                )
                if require_optimal_reactor:
                    raise ReactorAlreadyInstalledError()
        else:
            log.debug(
                'Running on "{platform}" and optimal reactor (Select) already installed.',
                platform=sys.platform,
            )

    from twisted.internet import reactor

    txaio.config.loop = reactor

    return reactor


def install_reactor(
    explicit_reactor=None, verbose=False, log=None, require_optimal_reactor=True
):
    """
    Install Twisted reactor.

    :param explicit_reactor: If provided, install this reactor. Else, install
        the optimal reactor.
    :type explicit_reactor: obj

    :param verbose: If ``True``, log (at level "info") the reactor that is
        in place afterwards.
    :type verbose: bool

    :param log: Explicit logging to this txaio logger object.
    :type log: obj

    :param require_optimal_reactor: If ``True`` and the desired reactor could not be installed,
        raise ``ReactorAlreadyInstalledError``, else fallback to another reactor.
    :type require_optimal_reactor: bool

    :returns: The Twisted reactor in place (`twisted.internet.reactor`).
    """
    if not log:
        log = txaio.make_logger()

    if explicit_reactor:
        # install explicitly given reactor
        #
        from twisted.application.reactors import installReactor

        if verbose:
            log.info(
                'Trying to install explicitly specified Twisted reactor "{reactor}" ..',
                reactor=explicit_reactor,
            )
        try:
            installReactor(explicit_reactor)
        except:
            log.failure(
                "Could not install Twisted reactor {reactor}\n{log_failure.value}",
                reactor=explicit_reactor,
            )
            sys.exit(1)
    else:
        # automatically choose optimal reactor
        #
        if verbose:
            log.info("Automatically choosing optimal Twisted reactor ..")
        install_optimal_reactor(require_optimal_reactor)

    # now the reactor is installed, import it
    from twisted.internet import reactor

    txaio.config.loop = reactor

    if verbose:
        from twisted.python.reflect import qual

        log.info(
            "Running on Twisted reactor {reactor}", reactor=qual(reactor.__class__)
        )

    return reactor
